import cudf
from typing import Literal


def print_text_red(text: str) -> str:
    return f"\033[1;31m{text}\033[0m"

def print_text_yellow(text: str) -> str:
    return f"\033[1;33m{text}\033[0m"

def print_text_green(text: str) -> str:
    return f"\033[1;32m{text}\033[0m"

def print_job_normalize_null_done() -> None:
    print(
        print_text_green("Done!"),
        "all",
        print_text_yellow("null"),
        "values converted to",
        print_text_yellow("<NA>")
    )

# RAW_INVALID_VALUES: list[str] = [
#     '',
#     ' ',
#     'NA',
#     '<NA>',
#     '(NA)',
#     'N/A',
#     'nan',
#     'NaT',
#     'NULL',
#     '(NULL)',
#     'NONE',
#     'NULO',
#     'NAN',
#     'UNDEFINED',
#     'None',
#     'NAO INFORMADO',
#     'NÃO INFORMADO',
#     'null',
#     '(null)',
#     'undefined',
#     'NOT AVAILABLE',
#     'NOT APPLICABLE',
#     'MISSING',
#     'UNKNOWN',
#     '-',
#     '--',
#     '?',
#     'N.D.',
# ]


RAW_INVALID_VALUES: list[str] = [
    '',
    ' ',
    '-',
    '--',
    '?',
]

RAW_INVALID_UPPERCASE_VALUES: list[str] = ['',
 'NA',
 '<NA>',
 '(NA)',
 'N/A',
 'NAN',
 'NAT',
 'NULL',
 '(NULL)',
 'NONE',
 'NULO',
 'NAN',
 'UNDEFINED',
 'NONE',
 'NAO INFORMADO',
 'NÃO INFORMADO',
 'NOT AVAILABLE',
 'NOT APPLICABLE',
 'MISSING',
 'UNKNOWN',
 'N.D.'
]


RAW_INVALID_LOWERRCASE_VALUES: list[str] = [
    'n.d.'
    'na',
    '<na>',
    '(na)',
    'n/a',
    'nan',
    'nat',
    'null',
    '(null)',
    'none',
    'nulo',
    'nan',
    'undefined',
     'none',
     'nao informado',
     'não informado',
     'undefined',
     'not available',
     'not applicable',
     'missing',
     'unknown'
]




# --- Constantes com os valores inválidos ---
# Mantidos fora da classe como constantes globais

RAW_INVALID_VALUES: list[str] = [
    '', ' ', '-', '--', '?',
]

RAW_INVALID_UPPERCASE_VALUES: list[str] = [
    'NA', '<NA>', '(NA)', 'N/A', 'NAN', 'NAT', 'NULL', '(NULL)', 'NONE',
    'NA/NA', 'NULO', 'UNDEFINED', 'NAO INFORMADO', 'NÃO INFORMADO',
    'NOT AVAILABLE', 'NOT APPLICABLE', 'MISSING', 'UNKNOWN', 'N.D.'
]

RAW_INVALID_LOWERRCASE_VALUES: list[str] = [
    'na', '<na>', '(na)', 'n/a', 'nan', 'nat', 'null', '(null)', 'none',
    'na/na', 'nulo', 'undefined', 'nao informado', 'não informado',
    'not available', 'not applicable', 'missing', 'unknown', 'n.d.'
]


# # class GPUNormalizeNullUtils:
#     """
#     Classe otimizada para normalizar valores nulos em colunas de um cudf.DataFrame
#     utilizando operações 100% na GPU.
#     """

class GPUNormalizeNullUtils:
    @staticmethod
    def normalize_na(
        current_df: cudf.DataFrame,
        additional_null_values: list[str] = [],
        # case_type: Literal['low', 'upper'] | None = None,
    ):
        # current_df = current_df.fillna(cudf.NA)

        for column_name in current_df.columns:
            GPUNormalizeNullUtils.normalize_na_column(
                current_df=current_df,
                column_name=column_name,
                additional_null_values=additional_null_values,
                # case_type=case_type
            )
        print_job_normalize_null_done()


    # 1. Pré-computa os valores e os armazena na GPU como atributos de classe
    # O uso de 'set' remove duplicatas antes de criar a Series
    _GPU_RAW = cudf.Series(list(set(RAW_INVALID_VALUES)))
    _GPU_UPPER = cudf.Series(list(set(RAW_INVALID_UPPERCASE_VALUES)))
    _GPU_LOWER = cudf.Series(list(set(RAW_INVALID_LOWERRCASE_VALUES)))
    _GPU_ALL = cudf.concat([_GPU_RAW, _GPU_UPPER, _GPU_LOWER]).unique()

    @staticmethod
    def normalize_na_column(
        current_df: cudf.DataFrame,
        column_name: str,
        additional_null_values: list[str] = [],
        case_type: Literal['low', 'upper'] | None = None,
    ) -> None:

        if column_name not in current_df.columns or current_df[column_name].empty:
            return

        if current_df[column_name].dtype not in ["object", "string"]:
            return
        
        if current_df[column_name].isnull().all():
            return

        # --- Seleção da lista de valores a serem substituídos (já na GPU) ---
        if case_type == 'low':
            values_to_replace = cudf.concat(
                [
                    GPUNormalizeNullUtils._GPU_RAW,
                    GPUNormalizeNullUtils._GPU_LOWER]
                )
        elif case_type == 'upper':
            values_to_replace = cudf.concat(
                [
                    GPUNormalizeNullUtils._GPU_RAW,
                    GPUNormalizeNullUtils._GPU_UPPER]
            )
        else:
            values_to_replace = GPUNormalizeNullUtils._GPU_ALL

        if additional_null_values:
            gpu_additional = cudf.Series(additional_null_values)
            values_to_replace = cudf.concat([values_to_replace, gpu_additional])

        values_to_replace = values_to_replace.unique()
        
        mask = current_df[column_name].isin(values_to_replace)

        current_df.loc[mask, column_name] = cudf.NA


    # @staticmethod
    # def normalize_na_column(
    #     current_df: cudf.DataFrame,
    #     column_name: str,
    #     additional_null_values: list[str] = []
    # ) -> None:        
    #     size: int = current_df[column_name].size

    #     # Coluna vazia
    #     if size == 0:
    #         return False
        
    #     # Coluna sem dados
    #     if current_df[column_name].notna().sum() == 0:
    #         return False

    #     # A coluna deve ser do tipo str
    #     if current_df[column_name].dtype not in ["object", "string"]:
    #         return False

    #     raw_invalid_values: list = RAW_INVALID_VALUES.copy()

    #     if additional_null_values:
    #         raw_invalid_values.extend(additional_null_values)
        
    #     current_df[column_name].replace(RAW_INVALID_VALUES, cudf.NA, inplace=True)




import cudf
import cupy as cp
from .gpu_analysis_utils import GPUAnalysisUtils

def print_text_red(text: str) -> str:
    return f"\033[1;31m{text}\033[0m"

def print_text_yellow(text: str) -> str:
    return f"\033[1;33m{text}\033[0m"

def print_text_green(text: str) -> str:
    return f"\033[1;32m{text}\033[0m"


def print_job_normalize_date_done() -> None:
    print(
        print_text_green("Done!"),
        "all columns containing",
        print_text_yellow("dates"),
        "values converted to",
        "have been normalized to ",
        print_text_yellow("YYYY-MM-DD")
    )


def print_job_normalize_datetime_column_done(
        column_name: str
    ) -> None:
        print(
            print_text_green("Done!"),
            "column",
            print_text_yellow(column_name),
            "containing",
            print_text_yellow("date"),
            "value",
            "have been normalized to",
            print_text_yellow("datedatetime64[ns]")
        )


def print_job_normalize_date_column_done(
        column_name: str
    ) -> None:
        print(
            print_text_green("Done!"),
            "column",
            print_text_yellow(column_name),
            "containing",
            print_text_yellow("date"),
            "value",
            "have been normalized to",
            print_text_yellow("yyyy-mm-dd")
        )


DATE_PATTERN = (
    r'^(?:'
    # Padrões com separadores: YYYY/MM/DD, DD/MM/YYYY, MM/DD/YYYY, etc.
    r'\d{4}[/\-._ ](0[1-9]|1[0-2])[/\-._ ](0[1-9]|[12]\d|3[01])|'
    r'(0[1-9]|[12]\d|3[01])[/\-._ ](0[1-9]|1[0-2])[/\-._ ]\d{4}|'
    r'(0[1-9]|1[0-2])[/\-._ ](0[1-9]|[12]\d|3[01])[/\-._ ]\d{4}|'
    r'(0[1-9]|[12]\d|3[01])[/\-._ ]\d{2}[/\-._ ]\d{2}|' # Formatos com ano de 2 dígitos
    r'\d{2}[/\-._ ](0[1-9]|[12]\d|3[01])[/\-._ ]\d{2}|'
    # Padrões sem separadores (8 dígitos)
    r'\d{4}(0[1-9]|1[0-2])(0[1-9]|[12]\d|3[01])|' # YYYYMMDD
    # r'(0[1-9]|1[0-2])(0[1-9]|[12]\d|3[01])\d{4}'   # MMDDYYYY
    r')$'
)


class GPUNormalizeDateUtils:
    @staticmethod
    def normalize_date(current_df: cudf.DataFrame) -> None:
        for column_name in current_df.columns:
            # if GPUNormalizeDateUtils.is_date(
            #     current_df,
            #     column_name
            # ):
            if GPUAnalysisUtils.infer_by_sample(
                series=current_df[column_name],
                regex_patern=DATE_PATTERN
            ):
                GPUNormalizeDateUtils.to_date_iso(
                    current_df=current_df,
                    column_name=column_name
                )
                print_job_normalize_date_column_done(column_name)

    @staticmethod
    def normalize_datetime(current_df: cudf.DataFrame) -> None:
        for column_name in current_df.columns:
            # if GPUNormalizeDateUtils.is_date(
            #     current_df,
            #     column_name
            # ):
            if GPUAnalysisUtils.infer_by_sample(
                series=current_df[column_name],
                regex_patern=DATE_PATTERN
            ):
                GPUNormalizeDateUtils.to_datetime(
                    current_df=current_df,
                    column_name=column_name
                )
                print_job_normalize_datetime_column_done(column_name)


    @staticmethod
    def is_date_column_by_samples(
        series: cudf.Series,
        n_parts: int = 10,
        n_samples: int = 10
    ) -> bool:
        """
        Verifica de forma performática na GPU se algum valor nas amostras de uma Series
        corresponde a um padrão de data.

        Args:
            s (cudf.Series): A Series de strings que será analisada.
            n_parts (int): O número de partes em que a Series será dividida para amostragem.
            n_samples (int): O número de amostras a serem coletadas de cada parte.

        Returns:
            bool: True se um padrão de data for encontrado nas amostras, False caso contrário.
        """

        series_size = len(series)

        # print(series.name)
        # return

        if ((n_parts * n_samples) >= series_size):
            raise ValueError("The total number of samples requested exceeds or equals the series size. Please provide a smaller value for n_parts or n_samples.")
        
        if (series_size // n_parts == 0):
            raise ValueError("The number of parts is greater than the series size. Please provide a smaller value for n_parts.")

        # Gera todos os índices de amostragem DE UMA VEZ na GPU
        step_pass = series_size // n_parts
        
        # Índices iniciais de cada bloco (ex: 0, 1000, 2000, ...)
        start_indices = cp.arange(n_parts) * step_pass
        
        # Offsets dentro de cada bloco (ex: 0, 1, 2, ... n_samples-1)
        sample_offsets = cp.arange(n_samples)

        all_indices = (start_indices[:, None] + sample_offsets).flatten()
        
        # Garante que os índices não ultrapassem o tamanho da Series
        all_indices = all_indices[all_indices < series_size]

        # 3. Seleção de todas as amostras em uma única operação
        samples = series.iloc[all_indices]

        date_pattern = (
            r'^(?:'
            # Padrões com separadores: YYYY/MM/DD, DD/MM/YYYY, MM/DD/YYYY, etc.
            r'\d{4}[/\-._ ](0[1-9]|1[0-2])[/\-._ ](0[1-9]|[12]\d|3[01])|'
            r'(0[1-9]|[12]\d|3[01])[/\-._ ](0[1-9]|1[0-2])[/\-._ ]\d{4}|'
            r'(0[1-9]|1[0-2])[/\-._ ](0[1-9]|[12]\d|3[01])[/\-._ ]\d{4}|'
            r'(0[1-9]|[12]\d|3[01])[/\-._ ]\d{2}[/\-._ ]\d{2}|' # Formatos com ano de 2 dígitos
            r'\d{2}[/\-._ ](0[1-9]|[12]\d|3[01])[/\-._ ]\d{2}|'
            # Padrões sem separadores (8 dígitos)
            r'\d{4}(0[1-9]|1[0-2])(0[1-9]|[12]\d|3[01])|' # YYYYMMDD
            # r'(0[1-9]|1[0-2])(0[1-9]|[12]\d|3[01])\d{4}'   # MMDDYYYY
            r')$'
        )

        if (samples.str.contains(date_pattern).sum()) == samples.size:
            return True
        return False


    @staticmethod
    def is_date(current_df: cudf.DataFrame, column_name:str) -> bool:
        if current_df.size == 0:
            return False

        # Coluna sem dados
        if current_df[column_name].notna().sum() == 0:
            return False

        if current_df[column_name].dtype not in ["object", "string"]:
            return False
        
        is_time = current_df[column_name].str.contains(":").any()

        if is_time:
            return False
    
        # yyyy?mm?dd
        mask = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 10) &
            (current_df[column_name].str.slice(0,4).str.isdigit()) &   # quatros digitos do ano
            (~current_df[column_name].str.slice(4,5).str.isdigit()) &  # separador: há algo que não é numero
            (~current_df[column_name].str.slice(4,5).str.isalpha()) &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(5,7).str.isdigit()) &   # dois digitos dos meses
            (~current_df[column_name].str.slice(7,8).str.isdigit()) &  # separador: há algo que não é numero
            (~current_df[column_name].str.slice(7,8).str.isalpha()) &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(8,10).str.isdigit())    # dois digitos dos dias
        )
        if mask.any():
            return True

        # dd?mm?yyyy
        mask = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 10) &
            (current_df[column_name].str.slice(0,2).str.isdigit()) &   # dois digitos dos dias
            (~current_df[column_name].str.slice(2,3).str.isdigit()) &  # separador: há algo que não é numero
            (~current_df[column_name].str.slice(2,3).str.isalpha()) &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(3,5).str.isdigit()) &   # dois digitos dos meses
            (~current_df[column_name].str.slice(5,6).str.isdigit()) &  # separador: há algo que não é numero
            (~current_df[column_name].str.slice(5,6).str.isalpha()) &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(6,10).str.isdigit())    # quatros digitos do ano
        )

        if mask.any():
            return True

        # dd?mm?yy
        mask = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 8) &
            (current_df[column_name].str.slice(0,2).str.isdigit()) &   # dois digitos dos dias
            (~current_df[column_name].str.slice(2,3).str.isdigit()) &  # separador: há algo que não é numero
            (~current_df[column_name].str.slice(2,3).str.isalpha()) &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(3,5).str.isdigit()) &   # dois digitos dos meses    
            (~current_df[column_name].str.slice(5,6).str.isdigit()) &  # separador: há algo que não é letra
            (~current_df[column_name].str.slice(5,6).str.isalpha()) &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(6,8).str.isdigit())     # dois digitos dos anos
        )
        if mask.any():
            return True
        
        days: list[str] = [str(number).zfill(2) for number in range(1, 32)]
        months: list[str] = [str(number).zfill(2) for number in range(1, 13)]

        mask = (
            (
                (current_df[column_name].notna()) &
                (current_df[column_name].str.len() == 8) &
                (current_df[column_name].str.isdigit()) &
                (current_df[column_name].str.slice(4, 6).isin(months)) &
                (current_df[column_name].str.slice(6, 8).isin(days))  
            )
            &
            (
                (current_df[column_name].notna()) &
                (current_df[column_name].str.isdigit())
            )
        )

        if mask.any():
            return True

        return False


    @staticmethod
    def infer_date_formats(current_df: cudf.DataFrame, column_name:str) -> list[tuple[str, int]]:

        if current_df.size == 0:
            return []

        # Coluna sem dados
        if current_df[column_name].notna().sum() == 0:
            return []

        if current_df[column_name].dtype not in ["object", "string"]:
            return []
        
        is_time = current_df[column_name].str.contains(":").any()

        if is_time:
            return []

        date_paterns: list[tuple[str, int]] = []
        
        valid_date_rows: int = 0

        # yyyy?mm?dd
        mask = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 10) &
            (current_df[column_name].str.slice(0,4).str.isdigit()) &   # quatros digitos do ano
            (~current_df[column_name].str.slice(4,5).str.isdigit()) &  # separador: há algo que não é numero
            (~current_df[column_name].str.slice(4,5).str.isalpha()) &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(5,7).str.isdigit()) &   # dois digitos dos meses
            (~current_df[column_name].str.slice(7,8).str.isdigit()) &  # separador: há algo que não é numero
            (~current_df[column_name].str.slice(7,8).str.isalpha()) &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(8,10).str.isdigit())    # dois digitos dos dias
        )
        
        if mask.any():
            valid_date_rows = valid_date_rows + mask.sum()
            date_paterns.append(
                ("%Y?%m?%d", mask.sum())
            )

        # dd?mm?yyyy
        mask = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 10) &
            (current_df[column_name].str.slice(0,2).str.isdigit()) &   # dois digitos dos dias
            (~current_df[column_name].str.slice(2,3).str.isdigit()) &  # separador: há algo que não é numero
            (~current_df[column_name].str.slice(2,3).str.isalpha()) &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(3,5).str.isdigit()) &   # dois digitos dos meses
            (~current_df[column_name].str.slice(5,6).str.isdigit()) &  # separador: há algo que não é numero
            (~current_df[column_name].str.slice(5,6).str.isalpha()) &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(6,10).str.isdigit())    # quatros digitos do ano
        )

        if mask.any():
            valid_date_rows = valid_date_rows + mask.sum()
            date_paterns.append(
                ("%m?%d?%Y", mask.sum())
            )       

        # dd?mm?yy
        mask = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 8) &
            (current_df[column_name].str.slice(0,2).str.isdigit()) &   # dois digitos dos dias
            (~current_df[column_name].str.slice(2,3).str.isdigit()) &  # separador: há algo que não é numero
            (~current_df[column_name].str.slice(2,3).str.isalpha()) &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(3,5).str.isdigit()) &   # dois digitos dos meses    
            (~current_df[column_name].str.slice(5,6).str.isdigit()) &  # separador: há algo que não é letra
            (~current_df[column_name].str.slice(5,6).str.isalpha()) &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(6,8).str.isdigit())     # dois digitos dos anos
        )
        if mask.any():
            valid_date_rows = valid_date_rows + mask.sum()
            date_paterns.append(
                ("%m?%d?%y", mask.sum())
            )

        days: list[str] = [str(number).zfill(2) for number in range(1, 32)]
        months: list[str] = [str(number).zfill(2) for number in range(1, 13)]

        mask = (
            (
                (current_df[column_name].notna()) &
                (current_df[column_name].str.len() == 8) &
                (current_df[column_name].str.isdigit()) &
                (current_df[column_name].str.slice(4, 6).isin(months)) &
                (current_df[column_name].str.slice(6, 8).isin(days))  
            )
            &
            (
                (current_df[column_name].notna()) &
                (current_df[column_name].str.isdigit())
            )
        )

        if mask.any():
            valid_date_rows = valid_date_rows + mask.sum()
            date_paterns.append(
                ("Y%m%d", mask.sum())
            )
            
        na_rows = (current_df[column_name].isna().sum())
        total_rows = current_df[column_name].size
        invalid_rows = total_rows - na_rows - valid_date_rows
        
        date_paterns.append(
            [
                ("<NA>", na_rows),
                ("invalid", invalid_rows)
            ]
        )

        return date_paterns
    

    @staticmethod
    def to_date_iso(current_df: cudf.DataFrame, column_name:str) -> None:
        if current_df.size == 0:
            return

        if current_df[column_name].notna().sum() == 0:
            return

        if current_df[column_name].dtype not in ["object", "string"]:
            return
        
        is_time = current_df[column_name].str.contains(":").any()
        
        if is_time:
            return

        current_df[column_name] = (
            current_df[column_name]
            .str.replace("/", "-", regex=False)
            .str.replace(" ", "-", regex=False)
            .str.replace("_", "-", regex=False)
            .str.replace(".", "-", regex=False)
        )

        # dd?mm?yyyy
        mask = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 10) &
            (current_df[column_name].str.slice(0,2).str.isdigit()) &        # dois digitos dos dias
            (current_df[column_name].str.slice(2,3) == "-") &               # separador: há algo que não é numero
            (current_df[column_name].str.slice(3,5).str.isdigit()) &        # dois digitos dos meses
            (current_df[column_name].str.slice(5,6) == "-") &               # separador: há algo que não é numero
            (current_df[column_name].str.slice(6,10).str.isdigit())         # quatros digitos do ano
        )

        if mask.any():
            current_df.loc[mask, column_name] = cudf.to_datetime(
                current_df.loc[mask, column_name], format="%d-%m-%Y"
            ).dt.strftime("%Y-%m-%d")
        del mask

        # dd?mm?yy
        mask = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 8) &
            (current_df[column_name].str.slice(0,2).str.isdigit()) &        # dois digitos dos dias
            (current_df[column_name].str.slice(2,3) == "-") &               # separador: há algo que não é numero
            (current_df[column_name].str.slice(3,5).str.isdigit()) &        # dois digitos dos meses    
            (current_df[column_name].str.slice(5,6) == "-") &               # separador: há algo que não é letra
            (current_df[column_name].str.slice(6,8).str.isdigit())          # dois digitos dos anos
        )

        if mask.any():
            current_df.loc[mask, column_name] = cudf.to_datetime(
                current_df.loc[mask, column_name], format="%d-%m-%y"
            ).dt.strftime("%Y-%m-%d")
        del mask

        days: list[str] = [str(number).zfill(2) for number in range(1, 32)]
        months: list[str] = [str(number).zfill(2) for number in range(1, 13)]
        
        # yyyymmdd
        mask = (
            (
                (current_df[column_name].notna()) &
                (current_df[column_name].str.len() == 8) &
                (current_df[column_name].str.isdigit()) &
                (current_df[column_name].str.slice(4, 6).isin(months)) &
                (current_df[column_name].str.slice(6, 8).isin(days))  
            )
            &
            (
                (current_df[column_name].notna()) &
                (current_df[column_name].str.isdigit())
            )
        )

        if mask.any():     
            current_df.loc[mask, column_name] = cudf.to_datetime(
                current_df.loc[mask, column_name], format='%Y%m%d'
            ).dt.strftime('%Y-%m-%d')
        del mask

        mask = (
            (   # yyyy?mm?ddd
                (current_df[column_name].notna()) &
                (current_df[column_name].str.len() == 10) &
                (current_df[column_name].str.slice(0,4).str.isdigit()) &   # quatros digitos do ano
                (~current_df[column_name].str.slice(4,5).str.isdigit()) &  # separador: há algo que não é numero
                (~current_df[column_name].str.slice(4,5).str.isalpha()) &  # separador: há algo que não é letra
                (current_df[column_name].str.slice(5,7).str.isdigit()) &   # dois digitos dos meses
                (~current_df[column_name].str.slice(7,8).str.isdigit()) &  # separador: há algo que não é numero
                (~current_df[column_name].str.slice(7,8).str.isalpha()) &  # separador: há algo que não é letra
                (current_df[column_name].str.slice(8,10).str.isdigit())    # dois digitos dos dias
            )
        )

        current_df.loc[~mask, column_name] = cudf.NA
        del mask

        return


    @staticmethod
    def to_datetime(current_df: cudf.DataFrame, column_name:str) -> None:
        if current_df.size == 0:
            return

        if current_df[column_name].notna().sum() == 0:
            return

        if current_df[column_name].dtype not in ["object", "string"]:
            return
        
        is_time = current_df[column_name].str.contains(":").any()
        
        if is_time:
            return

        current_df[column_name] = (
            current_df[column_name]
            .str.replace("/", "-", regex=False)
            .str.replace(" ", "-", regex=False)
            .str.replace("_", "-", regex=False)
            .str.replace(".", "-", regex=False)
        )

        # yyyy?mm?dd
        mask = (
                (current_df[column_name].notna()) &
                (current_df[column_name].str.len() == 10) &
                (current_df[column_name].str.slice(0,4).str.isdigit()) &    # quatros digitos do ano
                (current_df[column_name].str.slice(4,5) == "-") &           # separador: há algo que não é numero
                (current_df[column_name].str.slice(5,7).str.isdigit()) &    # dois digitos dos meses
                (current_df[column_name].str.slice(7,8) == "-") &           # separador: há algo que não é numero
                (current_df[column_name].str.slice(8,10).str.isdigit())     # dois digitos dos dias
        )

        if mask.any():
            current_df.loc[mask, column_name] = cudf.to_datetime(
                current_df.loc[mask, column_name], format="%Y-%m-%d"
            ).dt.strftime("%Y%m%d")
        del mask


        # dd?mm?yyyy
        mask = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 10) &
            (current_df[column_name].str.slice(0,2).str.isdigit()) &        # dois digitos dos dias
            (current_df[column_name].str.slice(2,3) == "-") &               # separador: há algo que não é numero
            (current_df[column_name].str.slice(3,5).str.isdigit()) &        # dois digitos dos meses
            (current_df[column_name].str.slice(5,6) == "-") &               # separador: há algo que não é numero
            (current_df[column_name].str.slice(6,10).str.isdigit())         # quatros digitos do ano
        )

        if mask.any():
            current_df.loc[mask, column_name] = cudf.to_datetime(
                current_df.loc[mask, column_name], format="%d-%m-%Y"
            ).dt.strftime("%Y%m%d")
        del mask


        # dd?mm?yy
        mask = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 8) &
            (current_df[column_name].str.slice(0,2).str.isdigit()) &        # dois digitos dos dias
            (current_df[column_name].str.slice(2,3) == "-") &               # separador: há algo que não é numero
            (current_df[column_name].str.slice(3,5).str.isdigit()) &        # dois digitos dos meses    
            (current_df[column_name].str.slice(5,6) == "-") &               # separador: há algo que não é letra
            (current_df[column_name].str.slice(6,8).str.isdigit())          # dois digitos dos anos
        )

        if mask.any():
            current_df.loc[mask, column_name] = cudf.to_datetime(
                current_df.loc[mask, column_name], format="%d-%m-%y"
            ).dt.strftime("%Y%m%d")
        del mask

        days: list[str] = [str(number).zfill(2) for number in range(1, 32)]
        months: list[str] = [str(number).zfill(2) for number in range(1, 13)]

        # yyyymmdd
        mask = (
            (
                (   
                    (current_df[column_name].notna()) &
                    (current_df[column_name].str.len() == 8) &
                    (current_df[column_name].str.isdigit()) &
                    (current_df[column_name].str.slice(4, 6).isin(months)) &
                    (current_df[column_name].str.slice(6, 8).isin(days))
                )
                &
                (
                    (current_df[column_name].notna()) &
                    (current_df[column_name].str.isdigit())
                )
                
            )
        )

        current_df.loc[~mask, column_name] = cudf.NA
        del mask

        current_df[column_name] = cudf.to_datetime(current_df[column_name], format='%Y%m%d')

        return


    @staticmethod
    def normalize_delimiter(current_df: cudf.DataFrame, column_name:str) -> None:
        if current_df.size == 0:
            return

        if current_df[column_name].notna().sum() == 0:
            return

        if current_df[column_name].dtype not in ["object", "string"]:
            return

        current_df[column_name] = (
            current_df[column_name]
            .str.replace("/", "-", regex=False)
            .str.replace(" ", "-", regex=False)
            .str.replace("_", "-", regex=False)
            .str.replace(".", "-", regex=False)
        )


    @staticmethod
    def convert_dd_mm_yyyy_to_yyyy_mm_dd(current_df: cudf.DataFrame, column_name:str) -> None:
        if current_df.size == 0:
            return

        if current_df[column_name].notna().sum() == 0:
            return

        if current_df[column_name].dtype not in ["object", "string"]:
            return
        
        mask_dd_mm_yyyy = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 10) &
            (current_df[column_name].str.slice(0,2).str.isdigit()) &   # dois digitos dos dias
            (current_df[column_name].str.slice(2,3) == "-") &  # separador: há algo que não é numero
            (current_df[column_name].str.slice(3,5).str.isdigit()) &    # dois digitos dos meses
            (current_df[column_name].str.slice(5,6) == "-") &  # separador: há algo que não é numero
            (current_df[column_name].str.slice(6,10).str.isdigit())   # quatros digitos do ano
        )

        if mask_dd_mm_yyyy.any():
            current_df.loc[mask_dd_mm_yyyy, column_name] = cudf.to_datetime(
                current_df.loc[mask_dd_mm_yyyy, column_name], format="%d-%m-%Y"
            ).dt.strftime("%Y-%m-%d")
        
        del mask_dd_mm_yyyy


    @staticmethod
    def convert_dd_mm_yy_to_yyyy_mm_dd(current_df: cudf.DataFrame, column_name:str) -> None:
        if current_df.size == 0:
            return

        if current_df[column_name].notna().sum() == 0:
            return

        if current_df[column_name].dtype not in ["object", "string"]:
            return
        
        mask_dd_mm_yy = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 8) &
            (current_df[column_name].str.slice(0,2).str.isdigit()) &   # dois digitos dos dias
            (current_df[column_name].str.slice(2,3) == "-") &  # separador: há algo que não é numero
            (current_df[column_name].str.slice(3,5).str.isdigit()) &    # dois digitos dos meses    
            (current_df[column_name].str.slice(5,6) == "-") &  # separador: há algo que não é letra
            (current_df[column_name].str.slice(6,8).str.isdigit())    # dois digitos dos anos
        )

        if mask_dd_mm_yy.any():
            current_df.loc[mask_dd_mm_yy, column_name] = cudf.to_datetime(
                current_df.loc[mask_dd_mm_yy, column_name], format="%d-%m-%y"
            ).dt.strftime("%Y-%m-%d")
            
        del mask_dd_mm_yy


    @staticmethod
    def convert_yyyyddmm_to_yyyy_mm_dd(current_df: cudf.DataFrame, column_name:str) -> None:
        if current_df.size == 0:
            return

        if current_df[column_name].notna().sum() == 0:
            return

        if current_df[column_name].dtype not in ["object", "string"]:
            return

        days: list[str] = [
            "01", "02", "03", "04", "05", "06", "07", "08", "09", "10",
            "11", "12", "13", "14", "15", "16", "17", "18", "19", "20",
            "21", "22", "23", "24", "25", "26", "27", "28", "29", "30",
            "31"
        ]
        
        months: list[str] = [
            "01", "02", "03", "04", "05", "06", "07", "08", "09", "10",
            "11", "12"
        ]
        
        # yyyymmdd
        mask_yyyymmdd = (
            (
                (current_df[column_name].notna()) &
                (current_df[column_name].str.len() == 8) &
                (current_df[column_name].str.isdigit()) &
                (current_df[column_name].str.slice(4, 6).isin(months)) &
                (current_df[column_name].str.slice(6, 8).isin(days))
            )
            &
            (
                (current_df[column_name].notna()) &
                (current_df[column_name].str.isdigit())
            )
        )
        
        if mask_yyyymmdd.any():     
            current_df.loc[mask_yyyymmdd, column_name] = cudf.to_datetime(
                current_df.loc[mask_yyyymmdd, column_name], format='%Y%m%d'
            ).dt.strftime('%Y-%m-%d')
            
        del mask_yyyymmdd


    @staticmethod
    def convert_invalid_date(current_df: cudf.DataFrame, column_name:str) -> None:
        if current_df.size == 0:
            return

        if current_df[column_name].notna().sum() == 0:
            return

        if current_df[column_name].dtype not in ["object", "string"]:
            return
        
        days: list[str] = [
            "01", "02", "03", "04", "05", "06", "07", "08", "09", "10",
            "11", "12", "13", "14", "15", "16", "17", "18", "19", "20",
            "21", "22", "23", "24", "25", "26", "27", "28", "29", "30",
            "31"
        ]
        
        months: list[str] = [
            "01", "02", "03", "04", "05", "06", "07", "08", "09", "10",
            "11", "12"
        ]
        
        all_dates_mask = (
            (   # yyyy?mm?dd
                (current_df[column_name].notna()) &
                (current_df[column_name].str.len() == 10) &
                (current_df[column_name].str.slice(0,4).str.isdigit()) &    # quatros digitos do ano
                (current_df[column_name].str.slice(4,5) == "-") &           # separador: há algo que não é numero
                (current_df[column_name].str.slice(5,7).str.isdigit()) &    # dois digitos dos meses
                (current_df[column_name].str.slice(7,8) == "-") &           # separador: há algo que não é numero
                (current_df[column_name].str.slice(8,10).str.isdigit())     # dois digitos dos dias
            )
            |
            (   # dd?mm?yyyy
                (current_df[column_name].notna()) &
                (current_df[column_name].str.len() == 10) &
                (current_df[column_name].str.slice(0,2).str.isdigit()) &    # dois digitos dos dias
                (current_df[column_name].str.slice(2,3) == "-") &           # separador: há algo que não é numero
                (current_df[column_name].str.slice(3,5).str.isdigit()) &    # dois digitos dos meses
                (current_df[column_name].str.slice(5,6) == "-") &           # separador: há algo que não é numero
                (current_df[column_name].str.slice(6,10).str.isdigit())     # quatros digitos do ano
            )
            |
            (
            # dd?mm?yy
            (current_df[column_name].notna()) &
            (current_df[column_name].str.len() == 8) &
            (current_df[column_name].str.slice(0,2).str.isdigit()) &        # dois digitos dos dias
            (current_df[column_name].str.slice(2,3) == "-") &               # separador: há algo que não é numero
            (current_df[column_name].str.slice(3,5).str.isdigit()) &        # dois digitos dos meses    
            (current_df[column_name].str.slice(5,6) == "-") &               # separador: há algo que não é letra
            (current_df[column_name].str.slice(6,8).str.isdigit())          # dois digitos dos anos
            )
            |
            (   # yyyymmdd
                (
                    (current_df[column_name].notna()) &
                    (current_df[column_name].str.len() == 8) &
                    (current_df[column_name].str.isdigit()) &
                    (current_df[column_name].str.slice(4, 6).isin(months)) &
                    (current_df[column_name].str.slice(6, 8).isin(days))
                )
                &
                (
                    (current_df[column_name].notna()) &
                    (current_df[column_name].str.isdigit())
                )
                
            )
        )

        current_df.loc[~all_dates_mask, column_name] = cudf.NA
            
        del all_dates_mask


    @staticmethod
    def convert_invalid_date_yyyy_mm_dd_to_na(current_df: cudf.DataFrame, column_name:str) -> None:
        if current_df.size == 0:
            return

        if current_df[column_name].notna().sum() == 0:
            return

        if current_df[column_name].dtype not in ["object", "string"]:
            return
        
        mask_yyyy_mm_dd = (
            (
                (current_df[column_name].notna()) &
                (current_df[column_name].str.len() == 10) &
                (current_df[column_name].str.slice(0,4).str.isdigit()) &    # quatros digitos do ano
                (current_df[column_name].str.slice(4,5) == "-") &           # separador: há algo que não é numero
                (current_df[column_name].str.slice(5,7).str.isdigit()) &    # dois digitos dos meses
                (current_df[column_name].str.slice(7,8) == "-") &           # separador: há algo que não é numero
                (current_df[column_name].str.slice(8,10).str.isdigit())     # dois digitos dos dias
            )
        )

        current_df.loc[~mask_yyyy_mm_dd, column_name] = cudf.NA
            
        del all_dates_mask

    
    @staticmethod
    def convert_invalid_date_yyyymmdd_to_na(current_df: cudf.DataFrame, column_name:str) -> None:
        if current_df.size == 0:
            return

        if current_df[column_name].notna().sum() == 0:
            return

        if current_df[column_name].dtype not in ["object", "string"]:
            return
        
        days: list[str] = [
            "01", "02", "03", "04", "05", "06", "07", "08", "09", "10",
            "11", "12", "13", "14", "15", "16", "17", "18", "19", "20",
            "21", "22", "23", "24", "25", "26", "27", "28", "29", "30",
            "31"
        ]
        
        months: list[str] = [
            "01", "02", "03", "04", "05", "06", "07", "08", "09", "10",
            "11", "12"
        ]
        
        mask_yyyymmdd = (
            (
                (
                    (current_df[column_name].notna()) &
                    (current_df[column_name].str.len() == 8) &
                    (current_df[column_name].str.isdigit()) &
                    (current_df[column_name].str.slice(4, 6).isin(months)) &
                    (current_df[column_name].str.slice(6, 8).isin(days))
                )
                &
                (
                    (current_df[column_name].notna()) &
                    (current_df[column_name].str.isdigit())
                )
                
            )
        )

        current_df.loc[~mask_yyyymmdd, column_name] = cudf.NA



DATE_REGEX_PATTERNS: list[tuple] = [
    # Formatos com barra
    (r'^\d{2}/\d{2}/\d{4}$', '%d/%m/%Y'),      # 01/02/2024
    (r'^\d{2}/\d{2}/\d{2}$', '%d/%m/%y'),      # 01/02/24
    (r'^\d{4}/\d{2}/\d{2}$', '%Y/%m/%d'),      # 2024/02/01
    (r'^\d{2}/\d{2}/\d{4} \d{2}:\d{2}$', '%d/%m/%Y %H:%M'),  # 01/02/2024 15:30
    (r'^\d{1,2}/\d{1,2}/\d{4}$', '%d/%m/%Y'),   # 1/1/2024
    (r'^\d{1,2}/\d{1,2}/\d{2}$', '%d/%m/%y'),   # 1/1/24

    # Formatos com hífen
    (r'^\d{2}-\d{2}-\d{4}$', '%d-%m-%Y'),      # 01-02-2024
    (r'^\d{2}-\d{2}-\d{2}$', '%d-%m-%y'),      # 01-02-24
    (r'^\d{4}-\d{2}-\d{2}$', '%Y-%m-%d'),      # 2024-02-01
    (r'^\d{5}-\d{2}-\d{2}$', '%Y-%m-%d'),      # 02024-02-01

    # Formatos com ponto
    (r'^\d{2}\.\d{2}\.\d{4}$', '%d.%m.%Y'),    # 01.02.2024
    (r'^\d{4}\.\d{2}\.\d{2}$', '%Y.%m.%d'),    # 2024.02.01

    # Numérico compacto
    (r'^\d{8}$', '%Y%m%d'),                    # 20240201

    # Formatos com nome de mês em inglês
    (r'^\d{2} [A-Za-z]{3} \d{4}$', '%d %b %Y'),    # 01 Feb 2024
    (r'^\d{2} [A-Za-z]+ \d{4}$', '%d %B %Y'),      # 01 February 2024
    (r'^[A-Za-z]{3} \d{2}, \d{4}$', '%b %d, %Y'),  # Feb 01, 2024
    (r'^[A-Za-z]+ \d{2}, \d{4}$', '%B %d, %Y'),    # February 01, 2024,
    (r'^\d{4} [A-Za-z]{3} \d{2}$', '%Y %b %d'),    # 2024 Feb 01
    (r'^\d{4} [A-Za-z]+ \d{2}$', '%Y %B %d'),      # 2024 February 01
]


TIME_REGEX_PATTERNS: list[tuple] = [
    (r'^\d{2}:\d{2}$', '%H:%M'),                # 23:45
    (r'^\d{2}:\d{2}:\d{2}$', '%H:%M:%S'),       # 23:45:12
    (r'^\d{2}:\d{2}:\d{2}\.\d+$', '%H:%M:%S.%f'), # 23:45:12.123456

    (r'^\d{1,2}:\d{2}\s?[AaPp][Mm]$', '%I:%M %p'),         # 3:45 PM
    (r'^\d{1,2}:\d{2}:\d{2}\s?[AaPp][Mm]$', '%I:%M:%S %p'), # 03:45:12 PM
    (r'^\d{1,2}:\d{2}:\d{2}\.\d+\s?[AaPp][Mm]$', '%I:%M:%S.%f %p'), # 3:45:12.123456 PM

    (r'^\d{4}\sUTC$', '%H%M UTC'),
]

DATE_TIME_REGEX_PATTERNS: list[tuple] = [
    # yyyy-mm-dd hh:mm:ss.sss
    (r'^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$', '%Y-%m-%d %H:%M:%S'),

    # yyyy-mm-dd hh:mm:ss.sss
    (r'^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{1,6}$', '%Y-%m-%d %H:%M:%S.%f'),
]

RAW_INVALID_VALUES: list[str] = [
    '',
    ' ',
    'NA',
    '<NA>',
    '(NA)',
    'N/A',
    'nan',
    'NaT',
    'NULL',
    '(NULL)',
    'NONE',
    'NULO',
    'NAN',
    'UNDEFINED',
    'None',
    'NAO INFORMADO',
    'NÃO INFORMADO',
    'null',
    '(null)',
    'undefined',
    'NOT AVAILABLE',
    'NOT APPLICABLE',
    'MISSING',
    'UNKNOWN',
    '-',
    '--',
    '?',
    'N.D.',
]

RAW_BOOLEAN_VALUES: list[tuple] = [
    ("True", "False"),       # Strings capitalizadas
    ("true", "false"),       # Strings minúsculas
    ("TRUE", "FALSE"),       # Strings maiúsculas
    ("yes", "no"),           # Configs / respostas curtas
    ("Yes", "No"),           # Capitalizadas
    ("YES", "NO"),           # Maiúsculas
    ("y", "n"),              # Abreviações minúsculas
    ("Y", "N"),              # Abreviações maiúsculas
    ("on", "off"),           # Configs de sistemas minúsculas
    ("On", "Off"),           # Capitalizadas
    ("ON", "OFF"),           # Maiúsculas
    ("1", "0"),              # Strings numéricas
    ("t", "f"),              # Abreviações PostgreSQL minúsculas
    ("T", "F")               # Abreviações PostgreSQL maiúsculas
]

SIGNED_INTEGER_TYPES: tuple[str] = (
    'int8',
    'int16',
    'int32',
    'int64'
)

UNSIGNED_INTEGER_TYPES: tuple[str] = (
    'uint32',
    'uint64'
)

FLOATING_TYPES: tuple[str] = (
    'float32',
    'float64'
)

NUMERIC_TYPES: tuple[str] = (
    'int8',
    'int16',
    'int32',
    'int64',
    'uint32',
    'uint64'
    'float32',
    'float64'
)


class RegexPatternsCombine:
    @staticmethod
    def combine_regex_patterns(
        date_regex_patterns: list[tuple[str, str]]
    ) -> str:
        """
        Recebe uma lista de tuplas (regex, formato) e retorna uma string única
        que une todos os regex com o operador OR (|), encapsulando cada regex
        em parênteses para evitar problemas de precedência.

        Exemplo de saída:
        r'(^\d{2}/\d{2}/\d{4}$)|(^\d{4}-\d{2}-\d{2}$)|...'
        """
        # Extrair os padrões regex da lista
        regex_patterns: list[str] = [pattern for pattern, _ in date_regex_patterns]

        grouped_patterns: list[str] = [f"({pattern})" for pattern in regex_patterns]

        combined_regex: str = "|".join(grouped_patterns)
        
        print(type(combined_regex))
        
        return combined_regex



from .gpu_normalize_string_utils import GPUNormalizeStringUtils
from .gpu_normalize_null_utils import GPUNormalizeNullUtils
from .gpu_analysis_utils import GPUAnalysisUtils
from .gpu_normalize_date_utils import GPUNormalizeDateUtils
from .gpu_normalize_time_utils import GPUNormalizeTimeUtils
from .gpu_normalize_datetime_utils import GPUNormalizeDateTimeUtils
from .gpu_normalize_boolean_utils import GPUNormalizeBooleanUtils
from .gpu_normalize_numeric_utils import GPUNormalizeNumericUtils
from .gpu_file_analysis_utils import GPUFileAnalysisUtils


class JiboiaGPU:
    # STRING
    normalize_str = staticmethod(GPUNormalizeStringUtils.normalize_str)
    to_category = staticmethod(GPUNormalizeStringUtils.to_category)
    normalize_category = staticmethod(GPUNormalizeStringUtils.normalize_category)
    normalize_unique_values = staticmethod(GPUNormalizeStringUtils.normalize_unique_values)
    
    # NULL
    normalize_na = staticmethod(GPUNormalizeNullUtils.normalize_na)

    # BOOL
    normalize_bool = staticmethod(GPUNormalizeBooleanUtils.normalize_bool)
    
    # NUMBER
    normalize_number = staticmethod(GPUNormalizeNumericUtils.normalize_number)
    is_probably_number = staticmethod(GPUNormalizeNumericUtils.is_probably_number)

    # DATE
    normalize_date = staticmethod(GPUNormalizeDateUtils.normalize_date)
    normalize_datetime = staticmethod(GPUNormalizeDateUtils.normalize_datetime)
    is_date_column_by_samples = staticmethod(GPUNormalizeDateUtils.is_date_column_by_samples)
    is_date = staticmethod(GPUNormalizeDateUtils.is_date)
    to_date_iso = staticmethod(GPUNormalizeDateUtils.to_date_iso)
    to_datetime = staticmethod(GPUNormalizeDateUtils.to_datetime)


    # TIME
    normalize_time = staticmethod(GPUNormalizeTimeUtils.normalize_time)
    is_time = staticmethod(GPUNormalizeTimeUtils.is_time)

    # TIMEDELTA
    normalize_timedelta = staticmethod(GPUNormalizeTimeUtils.normalize_timedelta)

    # DATETIME
    create_datetime_column_from_date_and_time = staticmethod(GPUNormalizeDateTimeUtils.create_datetime_column_from_date_and_time)
    create_day_week_column = staticmethod(GPUNormalizeDateTimeUtils.create_day_week_column)
    create_month_br_category_column = staticmethod(GPUNormalizeDateTimeUtils.create_month_br_category_column)
    create_day_week_br_category_column = staticmethod(GPUNormalizeDateTimeUtils.create_day_week_br_category_column)
    create_hour_category_column = staticmethod(GPUNormalizeDateTimeUtils.create_hour_category_column)

    # ANALYSIS
    drop_columns = staticmethod(GPUAnalysisUtils.drop_columns)
    cudf_size_info = staticmethod(GPUAnalysisUtils.cudf_size_info)
    pandasdf_size_info = staticmethod(GPUAnalysisUtils.pandasdf_size_info)
    get_gpu_memory_info = staticmethod(GPUAnalysisUtils.get_gpu_memory_info)
    is_vram_use_limit = staticmethod(GPUAnalysisUtils.is_vram_use_limit)
    get_index_samples = staticmethod(GPUAnalysisUtils.get_index_samples)
    infer_by_sample = staticmethod(GPUAnalysisUtils.infer_by_sample)

    # FILE
    get_csv_info = staticmethod(GPUFileAnalysisUtils.get_csv_info)
    convert_csv_to_utf8 = staticmethod(GPUFileAnalysisUtils.convert_csv_to_utf8)
    convert_all_csvs_to_utf8 = staticmethod(GPUFileAnalysisUtils.convert_all_csvs_to_utf8)
    read_csv_files = staticmethod(GPUFileAnalysisUtils.read_csv_files)

# instância única para ser usada como namespace
jiboia_gpu = JiboiaGPU()

__all__ = ["jiboia_gpu", "JiboiaGPU"]



import chardet
import csv
import cudf
import os
from pathlib import Path
from .gpu_analysis_utils import GPUAnalysisUtils


def print_text_red(text: str) -> str:
    return f"\033[1;31m{text}\033[0m"

def print_text_yellow(text: str) -> str:
    return f"\033[1;33m{text}\033[0m"

def print_text_green(text: str) -> str:
    return f"\033[1;32m{text}\033[0m"


class GPUFileAnalysisUtils:
    @staticmethod
    def get_csv_info(
        file_path: str,
        encoding: None|str = None,
        skiprows: None|int = 0,
        sample_characters: int = 2048,
        sample_bytes: int = 100000
    ) -> dict[str, any]:
        """
        Detecta o delimitador e o encoding de um arquivo CSV.

        :param file_path: caminho do arquivo CSV
        :param encoding: encoding forçado (se None, tenta detectar com chardet)
        :param skiprows: linhas iniciais a ignorar
        :param sample_characters: n de caracteres usados para sniffing
        :param sample_bytes: n de bytes usados para detectar encoding
        :return: (delimiter, encoding_detectado)
                :exemplo_de_retorno:
            {
                'encoding': 'utf-8',
                'confidence': 0.99,
                'language': 'pt',
                'delimiter': ','
            }
        """
        csv_info: dict[str, any] = {
            "encoding": encoding
        }

        # detecta encoding se não informado
        if encoding is None:
            with open(file_path, "rb") as f:
                rawdata = f.read(sample_bytes)
                enc_info = chardet.detect(rawdata)
                encoding = enc_info["encoding"] or "utf-8"

            csv_info = {**csv_info, **enc_info}

        # detecta delimitador
        with open(file_path, "r", encoding=encoding, errors="ignore") as file:
            for _ in range(skiprows):
                next(file, None)  # ignora linhas iniciais

            sample = file.read(sample_characters)
            sniffer = csv.Sniffer()
            try:
                dialect = sniffer.sniff(sample)
                delimiter = dialect.delimiter
            except csv.Error:
                delimiter = ";"

            csv_info["delimiter"] =  delimiter

        print(
            print_text_green("Done!"),
            "file:",
            print_text_yellow(file_path),
            "delimiter:",
            print_text_yellow(f'"{csv_info["delimiter"]}"'),
            "encoding:",
            print_text_yellow(csv_info["encoding"]),
        )
        return csv_info


    @staticmethod
    def convert_csv_to_utf8(file_path: str, output_path: str = None) -> str:
        """
        Converte um arquivo CSV para UTF-8.

        :param file_path: caminho do arquivo original
        :param output_path: caminho do arquivo convertido (se None, gera com _utf8.csv)
        :return: caminho do arquivo convertido em UTF-8
        """
        file_path = Path(file_path)

        # se não passar output_path, cria automático
        if output_path is None:
            output_path = file_path.with_name(file_path.stem + "_utf8.csv")

        # detecta encoding
        with open(file_path, "rb") as f:
            rawdata = f.read(100000)
            enc_info = chardet.detect(rawdata)
            source_encoding = enc_info["encoding"] or "latin-1"
            print(f"Encoding detectado: {source_encoding} (confiança={enc_info['confidence']:.2f})")

        # lê no encoding original e salva em UTF-8
        with open(file_path, "r", encoding=source_encoding, errors="ignore") as f_in:
            data = f_in.read()

        with open(output_path, "w", encoding="utf-8") as f_out:
            f_out.write(data)

        print(f"Arquivo convertido salvo em: {output_path}")
        return str(output_path)


    import chardet
    from pathlib import Path


    @staticmethod
    def convert_all_csvs_to_utf8(folder_path: str, new_folder: bool = False) -> None:
        """
        Converte todos os arquivos CSV em uma pasta para UTF-8.
        Detecta o encoding de cada arquivo automaticamente.
        
        :param folder_path: caminho da pasta contendo os CSVs
        :param new_folder: se True, cria uma subpasta 'utf8_converted', 
                           se False, salva os arquivos na mesma pasta com sufixo '_utf8'
        """
        folder = Path(folder_path)

        if new_folder:
            output_folder = folder / "utf8_converted"
            output_folder.mkdir(exist_ok=True)
        else:
            output_folder = folder  # salva na mesma pasta

        for csv_file in folder.glob("*.csv"):

            # Detecta encoding
            with open(csv_file, "rb") as f:
                rawdata = f.read(100000)  # lê parte do arquivo
                result = chardet.detect(rawdata)
                source_encoding = result["encoding"] or "latin-1"
                confidence = result["confidence"]
                # print(f"{csv_file.name}: encoding detectado = {source_encoding} (confiança={confidence:.2f})")

                print(
                    "file:",
                    csv_file,
                    "encoding:",
                    print_text_yellow(result["encoding"]),
                    "confidence:",
                    print_text_yellow(confidence),
                )

            # Se já estiver em UTF-8, pula
            if source_encoding.lower() == 'utf-8':
                continue

            # Lê e converte para UTF-8
            with open(csv_file, "r", encoding=source_encoding, errors="ignore") as f_in:
                data = f_in.read()

            if new_folder:
                output_file = output_folder / csv_file.name
            else:
                output_file = output_folder / f"{csv_file.stem}_utf8.csv"

            with open(output_file, "w", encoding="utf-8") as f_out:
                f_out.write(data)

            # print(f"Arquivo convertido salvo em: {output_file}")
            print(
                print_text_green("Done!"),
                "file:",
                print_text_yellow(output_file),
                "converted to",
                print_text_yellow("utf-8"),
            )


    @staticmethod
    def read_csv_files(
        folder_path: str,
        start_part: None|int = 1,
        end_part: None|int = None
    ) -> cudf.DataFrame:

        files_csv = sorted(
            [file for file in os.listdir(folder_path) if file.endswith(".csv")]
        )
        
        # Seleciona os arquivos baseado no indice
        start_idx = (start_part - 1) if start_part is not None else 0
        end_idx = end_part if end_part is not None else len(files_csv)
        
        selected_files = files_csv[start_idx:end_idx]
        
        df_cudf: cudf.DataFrame = cudf.DataFrame()
        
        for file_name in selected_files:
            file_path: str = f'{folder_path}{file_name}'
            
            csv_info: dict[str, any] = GPUFileAnalysisUtils.get_csv_info(
                file_path=file_path
            )
        
            sep_delimiter: str = csv_info["delimiter"]  
            
            if GPUAnalysisUtils.is_vram_use_limit():
                break
        
            df_cudf_part = cudf.read_csv(
                filepath_or_buffer=file_path,
                sep=sep_delimiter,
                dtype=str,
            )
        
            df_cudf = cudf.concat([df_cudf, df_cudf_part], ignore_index=True)
            
            GPUAnalysisUtils.cudf_size_info(df_cudf, print_info=True)
            del df_cudf_part
        
        return df_cudf



import cudf
import cupy as cp
from .gpu_analysis_utils import GPUAnalysisUtils


SIGNED_INTEGER_TYPES: tuple[str] = (
    'int8',
    'int16',
    'int32',
    'int64'
)

UNSIGNED_INTEGER_TYPES: tuple[str] = (
    'uint32',
    'uint64'
)

FLOATING_TYPES: tuple[str] = (
    'float32',
    'float64'
)

NUMERIC_TYPES: tuple[str] = (
    'int8',
    'int16',
    'int32',
    'int64',
    'uint32',
    'uint64'
    'float32',
    'float64'
)


NUMERIC_PATTERN = (
    r'^(?:'
    # === NÚMEROS ===
    r'-?\d+'                         # inteiro simples (ex.: 2003, -42)
    r'|'
    r'-?\d+(?:[.,]\d+)'              # decimal com ponto ou vírgula
    r'|'
    r'-?\d+(?:\.\d+)?[eE][+-]?\d+'   # notação científica
    r'|'
    r'-?\d{1,3}(?:\.\d{3})*,\d+'     # dd.ddd,dd
    r')$'
)


# mixed_pattern = (
#     r'^(?:'
#         # === 1. Inteiro simples ===
#     r'-?\d+'
#     r'|'
#         # === 2. Números com milhar e/ou decimal ===
#         # -?: opcional negativo
#         # (?:\d{1,3}(?:\.\d{3})+) : captura grupos de milhar (ex: 1.000, 5.500.000)
#         # |\d+ : ou inteiro simples sem milhar
#         # (?:[.,]\d+)? : decimal opcional (vírgula ou ponto)
#     r'-?(?:\d{1,3}(?:\.\d{3})+|\d+)(?:[.,]\d+)?'
#     r'|'
#         # === 3. Notação científica ===
#         # -?\d+ : parte inteira
#         # (?:\.\d+)? : decimal opcional
#         # [eE][+-]?\d+ : expoente obrigatório com E ou e, opcional + ou -
#     r'-?\d+(?:\.\d+)?[eE][+-]?\d+'
#     r')$'
# )

def print_text_red(text: str) -> str:
    return f"\033[1;31m{text}\033[0m"

def print_text_yellow(text: str) -> str:
    return f"\033[1;33m{text}\033[0m"

def print_text_green(text: str) -> str:
    return f"\033[1;32m{text}\033[0m"


def print_job_normalize_numeric_done(
        column_name: str,
        column_type: str
    ) -> None:
        print(
            print_text_green("Done!"),
            "column",
            print_text_yellow(column_name),
            "converted to",
            print_text_yellow(column_type)
        )


class GPUNormalizeNumericUtils:
    @staticmethod
    def normalize_number(
        current_df: cudf.DataFrame,
    ) -> bool:
        for column_name in current_df.columns:
            # Verifica se é um número por amostragem
            if GPUAnalysisUtils.infer_by_sample(
                series=current_df[column_name],
                regex_patern=NUMERIC_PATTERN,
                n_parts=100
            ):
                GPUNormalizeNumericUtils.normalize_number_format(
                    current_df=current_df,
                    column_name=column_name,
                )
                GPUNormalizeNumericUtils.convert_column_to_mumber(
                    current_df=current_df,
                    column_name=column_name,
                )

    @staticmethod
    def convert_column_to_mumber(
        current_df: cudf.DataFrame,
        column_name: str
    ) -> None:
        size: int = current_df[column_name].size

        # Coluna vazia
        if size == 0:
            return
        
        # Coluna sem dados
        if current_df[column_name].notna().sum() == 0:
            return
        
        # A coluna deve ser do tipo str
        if current_df[column_name].dtype not in ["object", "string"]:
            return

        is_number = current_df[column_name].str.isfloat()
        not_null = current_df[column_name].notna().sum()

        if (round((is_number.sum() / not_null) * 100) < 90):
            return

        if not is_number.any():
            return

        current_df.loc[~is_number, column_name] = cudf.NA
        
        del is_number

        current_df[column_name] = current_df[column_name].astype("float64")
        
        if ((current_df[column_name] % 1 == 0).all()):

            max_value = current_df[column_name].max()
            min_value = current_df[column_name].min()

            if max_value < 127 and min_value > -127:
                current_df[column_name] = current_df[column_name].astype('int8')
                print_job_normalize_numeric_done(column_name=column_name, column_type="int8")
                return

            if max_value < 32767 and min_value > -32768:
                current_df[column_name] = current_df[column_name].astype('int16')
                print_job_normalize_numeric_done(column_name=column_name, column_type="int16")
                return

            if max_value <= 2147483647 and min_value >= -2147483648:
                current_df[column_name] = current_df[column_name].astype('int32')
                print_job_normalize_numeric_done(column_name=column_name, column_type="int32")
                return

            current_df[column_name] = current_df[column_name].astype('int64')
            print_job_normalize_numeric_done(column_name=column_name, column_type="int64")
            return

        print_job_normalize_numeric_done(column_name=column_name, column_type="float64")
        return



# novos:
    @staticmethod
    def is_integer(
        current_series: cudf.Series,
        n_parts: int = 10,
        n_samples: int = 10
    ) -> bool:
        """
        Verifica de forma performática na GPU se algum valor nas amostras de uma Series
        corresponde a um padrão de data.

        Args:
            s (cudf.Series): A Series de strings que será analisada.
            n_parts (int): O número de partes em que a Series será dividida para amostragem.
            n_samples (int): O número de amostras a serem coletadas de cada parte.

        Returns:
            bool: True se um padrão de data for encontrado nas amostras, False caso contrário.
        """
        series_size = len(current_series)

        
        if (series_size <= 1000):
            int_pattern = (r'^-?\d+$')
            
            return bool(
                current_series
                .str.contains(int_pattern)
                .any()
            )

        if (n_parts * n_samples >= series_size):
            raise ValueError(
                "The total number of samples requested exceeds or equals the series size. Please provide a smaller value for n_parts or n_samples."
            )
        
        if (series_size // n_parts == 0):
            raise ValueError(
                "The number of parts is greater than the series size. Please provide a smaller value for n_parts."
            )
        
        # Índices iniciais de cada bloco (ex: 0, 1000, 2000, ...)
        start_indices = cp.arange(n_parts) * (cp.floor_divide(series_size, n_parts))
        
        # Offsets dentro de cada bloco (ex: 0, 1, 2, ... n_samples-1)
        sample_offsets = cp.arange(n_samples)

        all_indices = (start_indices[:, None] + sample_offsets).flatten()
        
        # Garante que os índices não ultrapassem o tamanho da Series
        all_indices = all_indices[all_indices < series_size]

        # 3. Seleção de todas as amostras em uma única operação
        current_series = current_series.iloc[all_indices]

        int_pattern = (r'^-?\d+$')

        # found_date_pattern = current_series.str.isinteger()
        found_date_pattern = current_series.str.contains(int_pattern)

        # O resultado de .any() é um único valor booleano que é transferido para a CPU
        return bool(found_date_pattern.notna().sum() == found_date_pattern.sum())


    @staticmethod
    def is_float(
        current_series: cudf.Series,
        n_parts: int = 10,
        n_samples: int = 10
    ) -> bool:
        series_size = len(current_series)

        if (series_size <= 1000):
            numeric_pattern = (r'^-?(\d+|\d+\.\d+)$')
            
            return bool(
                current_series
                .str.contains(numeric_pattern)
                .any()
            )

        if (n_parts * n_samples >= series_size):
            raise ValueError(
                "The total number of samples requested exceeds or equals the series size. Please provide a smaller value for n_parts or n_samples."
            )
        
        if (series_size // n_parts == 0):
            raise ValueError(
                "The number of parts is greater than the series size. Please provide a smaller value for n_parts."
            )
        
        # Índices iniciais de cada bloco (ex: 0, 1000, 2000, ...)
        start_indices = cp.arange(n_parts) * (cp.floor_divide(series_size, n_parts))
        
        # Offsets dentro de cada bloco (ex: 0, 1, 2, ... n_samples-1)
        sample_offsets = cp.arange(n_samples)

        all_indices = (start_indices[:, None] + sample_offsets).flatten()
        
        # Garante que os índices não ultrapassem o tamanho da Series
        all_indices = all_indices[all_indices < series_size]

        # 3. Seleção de todas as amostras em uma única operação
        current_series = current_series.iloc[all_indices]

        numeric_pattern = (r'^-?(\d+|\d+\.\d+)$')

        found_date_pattern = current_series.str.contains(numeric_pattern)

        return bool(
            found_date_pattern.notna().sum() == found_date_pattern.sum()
        )


    # def normalize_number(
    #         current_df: cudf.DataFrame,
    #         column_name:str
    # ) -> None:

    #     # ddd.ddd,dd
    #     mask = (
    #         (current_df[column_name].notna()) &
    #         (current_df[column_name].str.contains(".", regex=False)) &
    #         (current_df[column_name].str.contains(",", regex=False))
    #     )

    #     # print(current_df.loc[mask, column_name])

    #     current_df.loc[mask, column_name] = (
    #         current_df.loc[mask, column_name]
    #         .str.replace(".", "", regex=False)
    #         .str.replace(",", ".", regex=False)
    #     )
        
    #     # dd,dddd
    #     mask = (
    #         (current_df[column_name].notna()) &
    #         (current_df[column_name].str.contains(",")) &
    #         (~current_df[column_name].str.contains("."))
    #     )
        
    #     current_df.loc[mask, column_name] = (
    #         current_df.loc[mask, column_name].str.replace(",", ".", regex=False)
    #     )
    #     del mask


    @staticmethod
    def is_probably_number(
        current_df: cudf.DataFrame,
        column_name: str
    ) -> None:
        size: int = current_df[column_name].size

        # Coluna vazia
        if size == 0:
            return
        
        # Coluna sem dados
        if current_df[column_name].notna().sum() == 0:
            return
        
        # A coluna deve ser do tipo str
        if current_df[column_name].dtype not in ["object", "string"]:
            return

        # is_number = current_df[column_name].str.isfloat()
        # not_null = current_df[column_name].notna().sum()

        # print(column_name, "not_null :", not_null, "is_number", is_number.sum())

        # # Se os valores únicos são mais de 60% das linhas, não vale a pena
        # if (round((is_number.sum() / not_null) * 100) < 90):
            # print("não é numero: ", column_name)
            # return False
        # print("é numero: ", column_name)
        # print(column_name, is_number.sum())
        # print(is_number.sum())
        # return
        
        # if not is_number.any():
        #     return

        # current_df.loc[~is_number, column_name] = cudf.NA
        
        # del is_number

        # current_df[column_name] = current_df[column_name].astype("float64")
        
        if ((current_df[column_name] % 1 == 0).all()):

            max_value = current_df[column_name].max()
            min_value = current_df[column_name].min()

            if max_value < 127 and min_value > -127:
                current_df[column_name] = current_df[column_name].astype('int8')
                print_job_normalize_numeric_done(column_name=column_name, column_type="int8")
                return

            if max_value < 32767 and min_value > -32768:
                current_df[column_name] = current_df[column_name].astype('int16')
                print_job_normalize_numeric_done(column_name=column_name, column_type="int16")
                return

            if max_value <= 2147483647 and min_value >= -2147483648:
                current_df[column_name] = current_df[column_name].astype('int32')
                print_job_normalize_numeric_done(column_name=column_name, column_type="int32")
                return

            current_df[column_name] = current_df[column_name].astype('int64')
            print_job_normalize_numeric_done(column_name=column_name, column_type="int64")
            return

        print_job_normalize_numeric_done(column_name=column_name, column_type="float64")
        return


    @staticmethod
    def normalize_number_format(
        current_df: cudf.DataFrame,
        column_name:str
    ) -> None:
        
        # Coluna vazia
        if current_df[column_name].size == 0:
            return
        
        # Coluna sem dados
        if current_df[column_name].notna().sum() == 0:
            return
        
        # A coluna deve ser do tipo str
        if current_df[column_name].dtype not in ["object", "string"]:
            return

        # ddd.ddd,dd
        mask = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.contains(".", regex=False)) &
            (current_df[column_name].str.contains(",", regex=False))
        )

        current_df.loc[mask, column_name] = (
            current_df.loc[mask, column_name]
            .str.replace(".", "", regex=False)
            .str.replace(",", ".", regex=False)
        )
        
        # dd,dddd
        mask = (
            (current_df[column_name].notna()) &
            (current_df[column_name].str.contains(","))
        )
        
        current_df.loc[mask, column_name] = (
            current_df.loc[mask, column_name].str.replace(",", ".", regex=False)
        )
        del mask


import cudf


RAW_BOOLEAN_VALUES: list[tuple] = [
    ("True", "False"),       # Strings capitalizadas
    ("true", "false"),       # Strings minúsculas
    ("TRUE", "FALSE"),       # Strings maiúsculas
    ("yes", "no"),           # Configs / respostas curtas
    ("Yes", "No"),           # Capitalizadas
    ("YES", "NO"),           # Maiúsculas
    ("y", "n"),              # Abreviações minúsculas
    ("Y", "N"),              # Abreviações maiúsculas
    ("on", "off"),           # Configs de sistemas minúsculas
    ("On", "Off"),           # Capitalizadas
    ("ON", "OFF"),           # Maiúsculas
    ("1", "0"),              # Strings numéricas
    ("t", "f"),              # Abreviações PostgreSQL minúsculas
    ("T", "F")               # Abreviações PostgreSQL maiúsculas
]


BOOL_PATTERN = (
    r'^(?:'
    r'True|False|'
    r'true|false|'
    r'TRUE|FALSE|'
    r'yes|no|'
    r'Yes|No|'
    r'YES|NO|'
    r'y|n|'
    r'Y|N|'
    r'on|off|'
    r'On|Off|'
    r'ON|OFF|'
    r'1|0|'
    r't|f|'
    r'T|F'
    r')$'
)


def print_text_red(text: str) -> str:
    return f"\033[1;31m{text}\033[0m"

def print_text_yellow(text: str) -> str:
    return f"\033[1;33m{text}\033[0m"

def print_text_green(text: str) -> str:
    return f"\033[1;32m{text}\033[0m"


def print_job_normalize_bool_done(
        column_name: str,
        column_type: str
    ) -> None:
        print(
            print_text_green("Done!"),
            "column",
            print_text_yellow(column_name),
            "converted to",
            print_text_yellow(column_type)
        )


class GPUNormalizeBooleanUtils:
    @staticmethod
    def normalize_bool(current_df: cudf.DataFrame) -> None:
        for column_name in current_df.columns:
            GPUNormalizeBooleanUtils.convert_column_to_bool(
                current_df=current_df,
                column_name=column_name
            )

    @staticmethod
    def convert_column_to_bool(
        current_df: cudf.DataFrame,
        column_name: str
    ) -> bool:
        size: int = current_df[column_name].size

        # Coluna vazia
        if size == 0:
            return False

        # Coluna sem dados
        if current_df[column_name].notna().sum() == 0:
            return False

        if current_df[column_name].dtype not in ["object", "string"]:
            return False
        
        # A coluna não deve ser um tipo de booleano
        if current_df[column_name].dtype == 'bool':
            return False

        # Verifica as primeiras 10 primeiras linhas ou todo a coluna no caso de ser menor que 10
        rows_size_to_check: int = size if size <= 10 else 10

        preliminary_nor_null_rows_size: int = (
            current_df[column_name]
            .head(rows_size_to_check)
            .notna()
            .sum()
        )

        # Análise prévia da coluna considerando um número limitado de linhas
        for raw_boolean_value in RAW_BOOLEAN_VALUES:    
            preliminary_compatible_rows: int = (
                current_df[column_name]
                .dropna()
                .head(rows_size_to_check)
                .isin(raw_boolean_value)
                .sum()
            )

            # Se há indício de que pode ser compatível considerando as linhas analisadas
            if preliminary_compatible_rows == preliminary_nor_null_rows_size:

                # Análise profunda considerando todas as linhas
                notna_row_values: cudf.Series = (
                    current_df[column_name]
                    .dropna()
                    .isin(raw_boolean_value)
                )
                
                # Se todas as linhas não nulas são compatíveis então pode-se converter em bool
                if notna_row_values.sum() == current_df[column_name].notna().sum():
                    current_df[column_name] = (
                        current_df[column_name]
                        .map(
                            {
                                raw_boolean_value[0]: True,
                                raw_boolean_value[1]: False
                            }
                        )
                        .astype("boolean")
                    )
                    print_job_normalize_bool_done(column_name=column_name, column_type="boolean")



import cudf
import cupy as cp
import pandas as pd


def print_text_red(text: str) -> str:
    return f"\033[1;31m{text}\033[0m"

def print_text_yellow(text: str) -> str:
    return f"\033[1;33m{text}\033[0m"

def print_text_green(text: str) -> str:
    return f"\033[1;32m{text}\033[0m"


def print_job_drop_column_done(
    columns_to_delete: list[str]
) -> None:
    colored_names: list[str] = [
        print_text_yellow(name) for name in columns_to_delete
    ]
    
    msg: str = ", ".join(colored_names)

    print(
        print_text_green("Done!"),
        "column",
        msg,
        "was",
        print_text_red("dropped")
    )


class GPUAnalysisUtils:
    @staticmethod
    def drop_columns(
        current_df: cudf.DataFrame,
        column_names: list[str],
    ) -> None:

        columns_to_delete: list = []

        for column_name in column_names:

            if column_name in current_df.columns:
                columns_to_delete.append(column_name)


        if len(columns_to_delete) >= 1:

            current_df.drop(
                columns=columns_to_delete,
                inplace=True
            )
            print_job_drop_column_done(columns_to_delete)


    @staticmethod
    def cudf_size_info(current_df: cudf.DataFrame, print_info: bool = False) -> None:

        rows: int = current_df.shape[0]
        columns: int =  current_df.shape[1]
        vram_size_mb: float = round(current_df.memory_usage(index=True, deep=True).sum() / (1024 * 1024), 2)

        cudf_info: dict[str, any] = {
            "rows": rows,
            "columns": columns,
            "VRAM size Mb": vram_size_mb
        }

        if print_info:
            print(
                print_text_green("Done!"),
                "rows:",
                print_text_yellow(rows),
                "columns:",
                print_text_yellow(columns),
                "VRAM size Mb:",
                print_text_yellow(vram_size_mb),
            )
                
        return cudf_info


    @staticmethod
    def get_gpu_memory_info(device_id: int = 0) -> dict[str, int]:
        free_bytes, total_bytes = cp.cuda.runtime.memGetInfo()
        return {
            "free_mb": round(free_bytes / (1024 * 1024), 2),
            "total": round(total_bytes / (1024 * 1024), 2),
            "used_mb":  round((total_bytes - free_bytes) / (1024 * 1024), 2),
        }
    
    @staticmethod
    def is_vram_use_limit(device_id: int = 0) -> dict[str, int]:
        free_bytes, total_bytes = cp.cuda.runtime.memGetInfo()
        vram_percent_in_use: float = round(((total_bytes - free_bytes) / total_bytes) * 100, 1) >= 90
        
        if vram_percent_in_use >= 90:
            return True
        
        return False
    

    @staticmethod
    def pandasdf_size_info(current_df: pd.DataFrame) -> None:
        return {
            "rows": current_df.shape[0],
            "columns": current_df.shape[1],
            "RAM MB": round(current_df.memory_usage(index=True, deep=True).sum() / (1024 * 1024), 2)
        }
    

    def print_converted_column_type(
        column_name: str,
        column_type: str
    ) -> None:
        print(
            "\033[1;32mDone!\033[0m",
            "column",
            # f"{column_name}",
            f"\033[1;33m{column_name}\033[0m",
            "converted to", f"\033[1;33m{column_type}\033[0m"
        )
       
    # @staticmethod
    # def get_frequency_by_column(current_df: pd.DataFrame, column: str) -> pd.DataFrame:
    #     '''
    #     Gera um current_df com a frequencia de valores de uma coluna
    #     '''
    #     current_df_frequency: pd.DataFrame = (
    #         current_df[column]
    #         .value_counts()
    #         .reset_index()
    #     )
    
    #     current_df_frequency.columns = [column, 'frequency']
    
    #     current_df_frequency: pd.DataFrame = (
    #         current_df_frequency
    #         .sort_values(
    #             by='frequency',
    #             ascending=False
    #         )
    #     )
    
    #     return current_df_frequency


    # @staticmethod
    # def get_unique_values(current_df: pd.DataFrame, column: str) -> pd.DataFrame:
    #     '''
    #     Retorna um dataframe com os valores únicos de uma coluna
    #     '''
    #     df_unique = pd.DataFrame({
    #         column: current_df[column].dropna().unique()
    #     })

    #     df_unique = df_unique.sort_values(
    #         by=column,
    #         ascending=True
    #     ).reset_index(drop=True)

    #     return df_unique

    @staticmethod
    def get_index_samples(
        series: cudf.Series,
        n_parts: int = 10,
        n_samples: int = 10
    ) -> list[int]:
        """
        Verifica de forma performática na GPU se algum valor nas amostras de uma Series
        corresponde a um padrão de data.

        Args:
            s (cudf.Series): A Series de strings que será analisada.
            n_parts (int): O número de partes em que a Series será dividida para amostragem.
            n_samples (int): O número de amostras a serem coletadas de cada parte.

        Returns:
            bool: True se um padrão de data for encontrado nas amostras, False caso contrário.
        """

        series_size = len(series)

        # print(series.name)
        # return

        if ((n_parts * n_samples) >= series_size):
            raise ValueError("The total number of samples requested exceeds or equals the series size. Please provide a smaller value for n_parts or n_samples.")
        
        if (series_size // n_parts == 0):
            raise ValueError("The number of parts is greater than the series size. Please provide a smaller value for n_parts.")

        # Gera todos os índices de amostragem DE UMA VEZ na GPU
        step_pass = series_size // n_parts
        
        # Índices iniciais de cada bloco (ex: 0, 1000, 2000, ...)
        start_indices = cp.arange(n_parts) * step_pass
        
        # Offsets dentro de cada bloco (ex: 0, 1, 2, ... n_samples-1)
        sample_offsets = cp.arange(n_samples)

        all_indices = (start_indices[:, None] + sample_offsets).flatten()
        
        # Garante que os índices não ultrapassem o tamanho da Series
        all_indices = all_indices[all_indices < series_size]

        return all_indices
    

    @staticmethod
    def infer_by_sample(
        series: cudf.Series,
        regex_patern: str,
        n_parts: int = 10,
        n_samples: int = 10
    ) -> bool:
        """
        Verifica de forma performática na GPU se algum valor nas amostras de uma Series
        corresponde a um padrão de data.

        Args:
            s (cudf.Series): A Series de strings que será analisada.
            n_parts (int): O número de partes em que a Series será dividida para amostragem.
            n_samples (int): O número de amostras a serem coletadas de cada parte.

        Returns:
            bool: True se um padrão de data for encontrado nas amostras, False caso contrário.
        """
        series_size = len(series)

        if series_size == 0:
            return False

        # Coluna sem dados
        if series.notna().sum() == 0:
            return False

        if series.dtype not in ["object", "string"]:
            return False
        

        # print(series.name)
        # return

        if ((n_parts * n_samples) >= series_size):
            raise ValueError("The total number of samples requested exceeds or equals the series size. Please provide a smaller value for n_parts or n_samples.")
        
        if (series_size // n_parts == 0):
            raise ValueError("The number of parts is greater than the series size. Please provide a smaller value for n_parts.")

        # Gera todos os índices de amostragem DE UMA VEZ na GPU
        step_pass = series_size // n_parts
        
        # Índices iniciais de cada bloco (ex: 0, 1000, 2000, ...)
        start_indices = cp.arange(n_parts) * step_pass
        
        # Offsets dentro de cada bloco (ex: 0, 1, 2, ... n_samples-1)
        sample_offsets = cp.arange(n_samples)

        all_indices = (start_indices[:, None] + sample_offsets).flatten()
        
        # Garante que os índices não ultrapassem o tamanho da Series
        all_indices = all_indices[all_indices < series_size]

        # 3. Seleção de todas as amostras em uma única operação
        samples = series.iloc[all_indices]

        if (samples.str.contains(regex_patern).sum() == samples.notna().sum()):
            return True

        return False



import cudf
from .gpu_analysis_utils import GPUAnalysisUtils

def print_text_red(text: str) -> str:
    return f"\033[1;31m{text}\033[0m"

def print_text_yellow(text: str) -> str:
    return f"\033[1;33m{text}\033[0m"

def print_text_green(text: str) -> str:
    return f"\033[1;32m{text}\033[0m"


CUDF_TIME_REGEX_PATTERN: list[str] = [
    # 24h: HH:MM, HH:MM:SS, HH:MM:SS.sss... (fração opcional, 1–9 dígitos)
    r'^(?:[01]\d|2[0-3]):[0-5]\d(?:[:][0-5]\d(?:\.\d{1,9})?)?$',

    # 12h com AM/PM: H:MM, HH:MM, opcional :SS e .sss, AM/PM com ou sem pontos, qualquer caixa
    r'^(?:0?[1-9]|1[0-2]):[0-5]\d(?:[:][0-5]\d(?:\.\d{1,9})?)?\s*(?:[Aa]\.?[Mm]\.?|[Pp]\.?[Mm]\.?)$',

    # 12h somente hora inteira + AM/PM (ex.: "5 PM", "12am")
    r'^(?:0?[1-9]|1[0-2])\s*(?:[Aa]\.?[Mm]\.?|[Pp]\.?[Mm]\.?)$',

    # UTC: HHMM UTC
    r'^\d{4}\s*UTC$',
]


TIME_PATTERN = (
    r'^(?:'
    # hh:mm
    r'([01]\d|2[0-3]):[0-5]\d|'
    # hh:mm:ss
    r'([01]\d|2[0-3]):[0-5]\d:[0-5]\d|'
    # hhmm UTC
    r'([01]\d|2[0-3])[0-5]\d ?UTC|'
    # hhmmUTC
    r'([01]\d|2[0-3])[0-5]\dUTC'
    r')$'
)


def print_job_normalize_time_done() -> None:
    print(
        print_text_green("Done!"),
        "all columns containing",
        print_text_yellow("times"),
        "have been normalized to ",
        print_text_yellow("HH:MM:SS")
    )


def print_job_normalize_time_column_done(
        column_name: str
    ) -> None:
        print(
            print_text_green("Done!"),
            "column",
            print_text_yellow(column_name),
            "with",
            print_text_yellow("time"),
            "value",
            "have been normalized to",
            print_text_yellow("hh:mm:ss")
        )


def print_job_normalize_timedelta_column_done(
        column_name: str
    ) -> None:
        print(
            print_text_green("Done!"),
            "column",
            print_text_yellow(column_name),
            "with",
            print_text_yellow("time"),
            "value",
            "have been normalized to",
            print_text_yellow("timedelta64[s]")
        )


def print_job_create_category_column_done(
    column_name: str,
    column_from: str,
) -> None:
    print(
        print_text_green("Done!"),
        print_text_yellow("category"),
        "column",
        print_text_yellow(column_name),
        "created from",
        print_text_yellow(column_from),
        "type",
        print_text_yellow("category")
    )


class GPUNormalizeTimeUtils:
    @staticmethod
    def combine_regex_patterns(
        pattern_list: list[tuple[str, str, str]]
    ) -> str:
        """
        Combina múltiplos regex em um único padrão com OR (|),
        garantindo match da string inteira.
        """
        # extrai só o padrão da tupla
        patterns_only = [p[0] for p in pattern_list]
        # remove grupos de captura, senão vira confusão
        # transforma `(\d{4})-(\d{2})-(\d{2})` em `\d{4}-\d{2}-\d{2}`
        patterns_cleaned = [p.replace("(", "").replace(")", "") for p in patterns_only]

        # junta todos com OR
        combined_regex = r"^(?:{}$)".format("|".join(patterns_cleaned))

        return combined_regex


    @staticmethod
    def normalize_time(current_df: cudf.DataFrame) -> None:
        for column_name in current_df.columns:
            if GPUNormalizeTimeUtils.is_time(current_df, column_name):
                GPUNormalizeTimeUtils.convert_hhmm_utc_to_hh_mm_ss(
                    current_df=current_df,
                    column_name=column_name
                )
                GPUNormalizeTimeUtils.convert_hh_mm_to_hh_mm_ss(
                    current_df=current_df,
                    column_name=column_name
                )
                print_job_normalize_time_column_done(column_name)


    @staticmethod
    def normalize_timedelta(current_df: cudf.DataFrame) -> None:
        for column_name in current_df.columns:
            # if GPUNormalizeTimeUtils.is_time(current_df, column_name):
            if GPUAnalysisUtils.infer_by_sample(current_df[column_name], TIME_PATTERN):
                GPUNormalizeTimeUtils.convert_hhmm_utc_to_hh_mm_ss(
                    current_df=current_df,
                    column_name=column_name
                )
                GPUNormalizeTimeUtils.convert_hh_mm_to_hh_mm_ss(
                    current_df=current_df,
                    column_name=column_name
                )
                GPUNormalizeTimeUtils.to_timedelta_s(
                    current_df=current_df,
                    column_name=column_name
                )
                print_job_normalize_timedelta_column_done(column_name)


    @staticmethod
    def to_timedelta_s(
        current_df: cudf.DataFrame,
        column_name: str,
    ) -> None:

        current_df[column_name] = cudf.to_datetime(
            current_df[column_name], format="%H:%M:%S"
        )

        current_df[column_name] = current_df[column_name] - cudf.to_datetime(
            "00:00:00", format="%H:%M:%S"
        )

        current_df[column_name] = current_df[column_name].astype("timedelta64[s]")


    @staticmethod
    def normalize_time_to_timedelta_ns(
        current_df: cudf.DataFrame,
        column_name: str,
    ) -> None:
        
        size: int = current_df[column_name].size
        
        # A coluna deve ter pelo menos 1 linha
        if size == 0:
            return
        
        # A coluna não deve ser vazia
        if current_df[column_name].notna().sum() == 0:
            return

        # A coluna deve ser string
        if current_df[column_name].dtype not in ["object", "string"]:
            return

        # Quebra em horas, minutos e segundos
        split_values: cudf.DataFrame = (
            current_df[column_name]
            .str.split(":", expand=True)
        )
        split_values.columns = ["hours", "minutes", "seconds"]

        hours: cudf.Series = split_values["hours"].astype("int32")
        minutes: cudf.Series = split_values["minutes"].astype("int32")

        # Divide segundos em parte inteira e parte fracionária (se existir)
        sec_str: cudf.Series = split_values["seconds"]

        sec_int: cudf.Series = sec_str.str.extract(r"^(\d+)", expand=False).astype("int64")
        sec_frac: cudf.Series = (
            sec_str.str.extract(r"\.(\d+)", expand=False)
            .fillna("0")
        )

        # Normaliza fração para nanossegundos (preenche à direita até 9 dígitos)
        sec_frac_ns: cudf.Series = (
            sec_frac.str.pad(width=9, side="right", fillchar="0").astype("int64")
        )

        # Total em nanossegundos
        total_ns: cudf.Series = (
            (hours * 3600 + minutes * 60 + sec_int) * 1_000_000_000
            + sec_frac_ns
        )

        current_df[column_name] = total_ns.astype("timedelta64[ns]")


    # Normaliza UTC para HH:MM:SS
    @staticmethod
    def convert_hhmm_utc_to_hh_mm_ss(
            current_df: cudf.DataFrame,
            column_name: str
    ) -> None:
        mask_hh_mm_utc = (
            (current_df[column_name].notna())
            &
            (
                (current_df[column_name].str.len() == 8)
                |
                (current_df[column_name].str.len() == 7)
            )
            
            &
            (current_df[column_name].str.slice(0,4).str.isdigit())    # quatros digitos da hora
            &
            (current_df[column_name].str.contains("UTC"))   # separador: há algo que não é numero
        )

        current_df.loc[mask_hh_mm_utc, column_name] = (
            current_df.loc[mask_hh_mm_utc, column_name]
            .str.replace(" ", "")
            .str.replace("UTC", "")
        )
        
        current_df.loc[mask_hh_mm_utc, column_name] = (
            cudf.to_datetime(current_df.loc[mask_hh_mm_utc, column_name] + ":00", format="%H%M:%S")
            .dt.strftime("%H:%M:%S")
        )


    # Normaliza HH:MM:SS AM/PM para HH:MM:SS
    @staticmethod
    def normalize_column_time_am_pm(
        current_df: cudf.DataFrame,
        column_name: str,
    ) -> None:
        # Máscara para as linhas que contêm AM/PM
        mask = current_df[column_name].str.contains(r'(?:[Aa]\.?[Mm]\.?|[Pp]\.?[Mm]\.?)', regex=True)

        # Extrai grupos
        extracted = current_df.loc[mask, column_name].str.extract(
            r'^(\d{1,2}):(\d{2}):?(\d{2})?(?:\.\d+)?\s*([AaPp])\.?[Mm]\.?$',
            expand=True
        )
        
        extracted[2].replace('', '00', inplace=True)

        # Separa as colunas extraídas
        hours_am_pm = extracted[0].astype("int32")
        minutes = extracted[1]
        seconds = extracted[2]
        am_pm = extracted[3]

        
        # Lógica de conversão de AM/PM para 24h
        is_pm_mask = (am_pm.str.upper() == 'P') & (hours_am_pm != 12)
        is_12am_mask = (am_pm.str.upper() == 'A') & (hours_am_pm == 12)

        # Cria uma cópia das horas para evitar alterar a série original
        hours_24h = hours_am_pm.copy()
        hours_24h.loc[is_pm_mask] = hours_24h + 12
        hours_24h.loc[is_12am_mask] = 0

        # Concatena as partes para o formato 'HH:MM:SS'
        normalized = (
            hours_24h.astype("str").str.zfill(2)
            + ":" + minutes + ":" + seconds
        )

        # Altera a coluna original apenas onde a máscara é True
        current_df.loc[mask, column_name] = normalized


    # Normaliza HH:MM para HH:MM:SS
    @staticmethod
    def convert_hh_mm_to_hh_mm_ss(
            current_df: cudf.DataFrame,
            column_name: str
    ) -> None:

        mask_hh_mm = (
            (current_df[column_name].notna())
            &
            (current_df[column_name].str.len() == 5)
            &
            (current_df[column_name].str.slice(0,2).str.isdigit())    # quatros digitos da hora
            &
            (current_df[column_name].str.slice(2,3) == ":")   # separador: há algo que não é numero
            &
            (current_df[column_name].str.slice(3,5).str.isdigit())    # dois digitos dos minutos
        )
        
        current_df.loc[mask_hh_mm, column_name] = current_df.loc[mask_hh_mm, column_name] + ":00"


    # Novo:
    @staticmethod
    def is_time(current_df: cudf.DataFrame, column_name:str) -> bool:
        if current_df.size == 0:
            return False

        # Coluna sem dados
        if current_df[column_name].notna().sum() == 0:
            return False

        if current_df[column_name].dtype not in ["object", "string"]:
            return False
        
        not_na_rows = current_df[column_name].notna().sum()

        mask_not_time = (
            (current_df[column_name].notna())
            &
            (~current_df[column_name].str.slice(0, 1).str.isdigit())
            &
            (
                (~current_df[column_name].str.contains("UTC"))
                |
                (~current_df[column_name].str.contains(":"))
            )
        )

        if mask_not_time.sum() == not_na_rows:
            return False
        del mask_not_time

        mask_hh_mm_ss_decimal = (
            (current_df[column_name].notna())
            &
            (current_df[column_name].str.len() > 8)
            &
            (current_df[column_name].str.slice(0, 2).str.isdigit())  # HH
            &
            (current_df[column_name].str.slice(2, 3) == ":")   # :
            &
            (current_df[column_name].str.slice(3, 5).str.isdigit())  # MM
            &
            (current_df[column_name].str.slice(5, 6) == ":")   # :
            &
            (current_df[column_name].str.slice(6, 8).str.isdigit())  # SS
            &
            (current_df[column_name].str.slice(8, 9) == ".")   # .
        )

        mask_hh_mm_ss = (
            (current_df[column_name].notna())
            &
            (current_df[column_name].str.len() == 8)
            &
            (current_df[column_name].str.slice(0,2).str.isdigit())    # quatros digitos da hora
            &
            (current_df[column_name].str.slice(2,3) == ":")   # separador: há algo que não é numero
            &
            (current_df[column_name].str.slice(3,5).str.isdigit())    # dois digitos dos minutos
            &
            (current_df[column_name].str.slice(5,6) == ":")   # separador: há algo que não é numero
            &
            (current_df[column_name].str.slice(6,8).str.isdigit())   # dois digitos dos segundos
        )
        
        mask_hh_mm = (
            (current_df[column_name].notna())
            &
            (current_df[column_name].str.len() == 5)
            &
            (current_df[column_name].str.slice(0,2).str.isdigit())    # quatros digitos da hora
            &
            (current_df[column_name].str.slice(2,3) == ":")   # separador: há algo que não é numero
            &
            (current_df[column_name].str.slice(3,5).str.isdigit())    # dois digitos dos minutos
        )
        
        mask_hh_mm_utc = (
            (current_df[column_name].notna())
            &
            (
                (current_df[column_name].str.len() == 8)
                |
                (current_df[column_name].str.len() == 7)
            )
            
            &
            (current_df[column_name].str.slice(0,4).str.isdigit())    # quatros digitos da hora
            &
            (current_df[column_name].str.contains("UTC"))   # separador: há algo que não é numero
        )
        
        is_time = mask_hh_mm_ss_decimal | mask_hh_mm_ss | mask_hh_mm | mask_hh_mm_utc

        if round((is_time.sum() / not_na_rows) * 100) >= 50:
            return True
        return False


import cudf
from typing import Literal
from .gpu_analysis_utils import GPUAnalysisUtils


def print_text_red(text: str) -> str:
    return f"\033[1;31m{text}\033[0m"

def print_text_yellow(text: str) -> str:
    return f"\033[1;33m{text}\033[0m"

def print_text_green(text: str) -> str:
    return f"\033[1;32m{text}\033[0m"


MIXED_PATTERN = (
    r'^(?:'
    # === NÚMEROS ===
    r'-?\d+'                         # inteiro simples (ex.: 2003, -42)
    r'|'
    r'-?\d+(?:[.,]\d+)'              # decimal com ponto ou vírgula
    r'|'
    r'-?\d+(?:\.\d+)?[eE][+-]?\d+'   # notação científica
    r'|'
    r'-?\d{1,3}(?:\.\d{3})*,\d+'     # dd.ddd,dd
    r'|'
    # === DATAS ===
    r'\d{4}[/\-._ ](0[1-9]|1[0-2])[/\-._ ](0[1-9]|[12]\d|3[01])'  # YYYY/MM/DD
    r'|'
    r'(0[1-9]|[12]\d|3[01])[/\-._ ](0[1-9]|1[0-2])[/\-._ ]\d{4}'  # DD/MM/YYYY
    r'|'
    r'(0[1-9]|1[0-2])[/\-._ ](0[1-9]|[12]\d|3[01])[/\-._ ]\d{4}'  # MM/DD/YYYY
    r'|'
    r'(0[1-9]|[12]\d|3[01])[/\-._ ]\d{2}[/\-._ ]\d{2}'            # DD/MM/YY
    r'|'
    r'\d{2}[/\-._ ](0[1-9]|[12]\d|3[01])[/\-._ ]\d{2}'            # YY/MM/DD
    r'|'
    r'\d{4}(0[1-9]|1[0-2])(0[1-9]|[12]\d|3[01])'                  # YYYYMMDD
    r'|'
    # === HORAS ===
    r'([01]\d|2[0-3]):[0-5]\d'            # hh:mm
    r'|'
    r'([01]\d|2[0-3]):[0-5]\d:[0-5]\d'    # hh:mm:ss
    r'|'
    r'([01]\d|2[0-3])[0-5]\d ?UTC'        # hhmm UTC
    r'|'
    r'([01]\d|2[0-3])[0-5]\dUTC'          # hhmmUTC
    r'|'
    # === BOOLEANOS ===
    r'True|False|'
    r'true|false|'
    r'TRUE|FALSE|'
    r'yes|no|'
    r'Yes|No|'
    r'YES|NO|'
    r'y|n|'
    r'Y|N|'
    r'on|off|'
    r'On|Off|'
    r'ON|OFF|'
    r't|f|'
    r'T|F'
    r')$'
)



def print_job_normalize_string_done(
    to_case: None|str,
    to_ASCII: bool = False
) -> None:

    to_case_msg: str = ''

    if to_case == "upper":
        to_case_msg = "uppercase"
    if to_case == "lower":
        to_case_msg = "lowercase"

    to_ASCII_msg: str = "ASCII" if to_ASCII else ""

    msg: str = " ".join(filter(None, [to_case_msg, to_ASCII_msg]))

    # msg: str = "to_case_msg"

    if len(msg) > 0:
        print(
            print_text_green("Done!"),
            "all",
            print_text_yellow("string"),
            "were converted to",
            print_text_yellow(msg)
        )


def print_job_normalize_string_to_ascii_done(
    case_type: None|str = None,
) -> None:
    
    case_type_msg: str = ''

    if case_type == "upper":
        case_type_msg = "uppercase"
    if case_type == "low":
        case_type_msg = "lowercase"

    to_ASCII_msg: str = "ASCII"

    msg: str = " ".join(filter(None, [to_ASCII_msg, case_type_msg]))

    # msg: str = "to_ASCII_msg"

    if len(msg) > 0:
        print(
            print_text_green("Done!"),
            "all",
            print_text_yellow("string"),
            "were converted to",
            print_text_yellow(msg)
        )


def print_job_normalize_space_done() -> None:
    print(
        print_text_green("Done!"),
        "all duplicate and edge",
        print_text_yellow("spaces"),
        "have been",
        print_text_yellow("removed")
    )


def print_job_create_category_done(
    column_name: str,
) -> None:

    print(
        print_text_green("Done!"),
        "the column",
        print_text_yellow(column_name),
        "was converted to a",
        print_text_yellow("category")
    )



class GPUNormalizeStringUtils:
    # == CONVERTE TODAS AS COLUNAS EM STRING == #
    @staticmethod
    def convert_df_to_string(current_df: cudf.DataFrame):
        for column in current_df.columns:
            current_df[column] = (
                current_df[column]
                    .astype(str)
            )
        print('columns with boolean types have been normalized')


    # == CRIA CATEGORIAS DE COLUNAS DE STRING, SE FOR POSSÍVEL == #
    @staticmethod
    def normalize_category(current_df: cudf.DataFrame):
        for column_name in current_df.columns:
            # Verifica se a coluna é do tipo string
            if not GPUNormalizeStringUtils.is_string(current_df, column_name):
                continue
            # Verifica se mesmo sendo string, é uma coluna de outro tipo de dados como int, bool, float
            if not GPUNormalizeStringUtils.is_probably_text(current_df, column_name):
                continue
            GPUNormalizeStringUtils.to_category(
                current_df,
                column_name
            )


    # == REMOVE TODOS OS ESPAÇOS DUPLOS, NO COMEÇO E FINAL DE STRING == #
    @staticmethod
    def normalize_spaces_df(current_df: cudf.DataFrame) -> None:
        for column_name in current_df.columns:
            size: int = current_df[column_name].size

            # Coluna vazia
            if size == 0:
                continue

            # Coluna sem dados
            if current_df[column_name].notna().sum() == 0:
                continue

            # A coluna deve ser do tipo str
            if current_df[column_name].dtype not in ["object", "string"]:
                continue

            current_df[column_name] = (
                current_df[column_name]
                .str.normalize_spaces()
                .str.strip()
            )
        print_job_normalize_space_done()


    # == NORMALIZA TODAS AS STRINGS DO DF == #
    @staticmethod
    def normalize_str(
        current_df: cudf.DataFrame,
        to_case: Literal['lower', 'upper'] | None = None,
        to_ASCII: bool=False,
    ) -> None:
        for column_name in current_df.columns:
            if column_name not in current_df.columns or current_df[column_name].empty:
                continue

            if current_df[column_name].dtype not in ["object", "string"]:
                continue
            
            if current_df[column_name].isnull().all():
                continue
            
            processed_column = (
                current_df[column_name]
                .str.normalize_spaces()
                .str.strip()
            )               

            if to_case == "low":
                processed_column = (
                    processed_column
                    .str.lower()
                )

                if to_ASCII:
                    processed_column = (
                    processed_column
                        .str.replace(r"[áàâãä]", "a", regex=True)
                        .str.replace(r"[éèêë]", "e", regex=True)
                        .str.replace(r"[íìîï]", "i", regex=True)
                        .str.replace(r"[óòôõö]", "o", regex=True)
                        .str.replace(r"[úùûü]", "u", regex=True)
                        .str.replace(r"[ç]", "c", regex=True)
                )

            if to_case == "upper":
                processed_column = (
                    processed_column
                    .str.upper()
                )
                if to_ASCII:
                    processed_column = (
                        processed_column
                        .str.replace(r"[ÁÀÂÃÄ]", "A", regex=True)
                        .str.replace(r"[ÉÈÊË]", "E", regex=True)
                        .str.replace(r"[ÍÌÎÏ]", "I", regex=True)
                        .str.replace(r"[ÓÒÔÕÖ]", "O", regex=True)
                        .str.replace(r"[ÚÙÛÜ]", "U", regex=True)
                        .str.replace(r"[Ç]", "C", regex=True)
                    )

            if to_ASCII and not to_case:
                processed_column = (
                    processed_column
                        .str.replace(r"[áàâãä]", "a", regex=True)
                        .str.replace(r"[ÁÀÂÃÄ]", "A", regex=True)
                        .str.replace(r"[éèêë]", "e", regex=True)
                        .str.replace(r"[ÉÈÊË]", "E", regex=True)
                        .str.replace(r"[íìîï]", "i", regex=True)
                        .str.replace(r"[ÍÌÎÏ]", "I", regex=True)
                        .str.replace(r"[óòôõö]", "o", regex=True)
                        .str.replace(r"[ÓÒÔÕÖ]", "O", regex=True)
                        .str.replace(r"[úùûü]", "u", regex=True)
                        .str.replace(r"[ÚÙÛÜ]", "U", regex=True)
                        .str.replace(r"[ç]", "c", regex=True)
                        .str.replace(r"[Ç]", "C", regex=True)
                    )

            current_df[column_name] = processed_column

        print_job_normalize_space_done()
        if to_case or to_ASCII:
            print_job_normalize_string_done(to_case=to_case, to_ASCII=to_ASCII)


    @staticmethod
    def to_ascii(
        current_df: cudf.DataFrame,
        case_type: None|str = None
    ) -> None:
        for column_name in current_df.columns:
            size: int = current_df[column_name].size

            if size == 0:
                print("empty column")
                continue

            if current_df[column_name].dtype not in ["object", "string"]:
                continue
            
            if not case_type:
                current_df[column_name] = (
                    current_df[column_name]
                    .str.replace(r"[áàâãä]", "a", regex=True)
                    .str.replace(r"[ÁÀÂÃÄ]", "A", regex=True)
                    .str.replace(r"[éèêë]", "e", regex=True)
                    .str.replace(r"[ÉÈÊË]", "E", regex=True)
                    .str.replace(r"[íìîï]", "i", regex=True)
                    .str.replace(r"[ÍÌÎÏ]", "I", regex=True)
                    .str.replace(r"[óòôõö]", "o", regex=True)
                    .str.replace(r"[ÓÒÔÕÖ]", "O", regex=True)
                    .str.replace(r"[úùûü]", "u", regex=True)
                    .str.replace(r"[ÚÙÛÜ]", "U", regex=True)
                    .str.replace(r"[ç]", "c", regex=True)
                    .str.replace(r"[Ç]", "C", regex=True)
                )
                continue

            if case_type == 'lower':
                current_df[column_name] = (
                    current_df[column_name]
                    .str.replace(r"[áàâãä]", "a", regex=True)
                    .str.replace(r"[éèêë]", "e", regex=True)
                    .str.replace(r"[íìîï]", "i", regex=True)
                    .str.replace(r"[óòôõö]", "o", regex=True)
                    .str.replace(r"[úùûü]", "u", regex=True)
                    .str.replace(r"[ç]", "c", regex=True)
                )
                continue

            if case_type == 'upper':
                current_df[column_name] = (
                    current_df[column_name]
                    .str.replace(r"[ÁÀÂÃÄ]", "A", regex=True)
                    .str.replace(r"[ÉÈÊË]", "E", regex=True)
                    .str.replace(r"[ÍÌÎÏ]", "I", regex=True)
                    .str.replace(r"[ÓÒÔÕÖ]", "O", regex=True)
                    .str.replace(r"[ÚÙÛÜ]", "U", regex=True)
                    .str.replace(r"[Ç]", "C", regex=True)
                )
                continue

        print_job_normalize_string_to_ascii_done(case_type=case_type)
        return

    @staticmethod
    def to_upercase(
        current_df: cudf.DataFrame,
    ) -> None:
        for column_name in current_df.columns:
            size: int = current_df[column_name].size

            if size == 0:
                continue

            if current_df[column_name].dtype in ["object", "string"]:

                current_df[column_name] = (
                    current_df[column_name]
                    .str.upper()
                )
                continue
        
        print_job_normalize_string_done(to_case='upper')


    @staticmethod
    def normalize_unique_values_in_column(
        current_df: cudf.DataFrame, 
        column_name: str, 
        target_value: str, 
        values_to_replace: list[str]
    ) -> None:
        '''
        Substitui valores específicos de uma coluna por um valor alvo.
        
        Parâmetros:
        - current_df: DataFrame contendo os dados.
        - column: Nome da coluna onde a substituição ocorrerá.
        - target_value: O valor que substituirá os valores encontrados.
        - values_to_replace: Lista de valores que serão substituídos pelo valor alvo.
        '''
        size: int = current_df[column_name].size
        if size == 0:
            print("empty column")
            return

        if current_df[column_name].dtype not in ["object", "string"]:
            return

        # Substitui os valores encontrados pela coluna `column` que estão na lista `values_to_replace` com `target_value`
        current_df[column_name].replace(
            values_to_replace,
            target_value, 
            inplace=True
        )


    @staticmethod
    def is_category(
        current_df: cudf.DataFrame,
        column_name: str
    ) -> None:
        # Coluna sem dados
        if current_df[column_name].notna().sum() == 0:
            return False
        
        # A coluna deve ser do tipo str
        if current_df[column_name].dtype not in ["object", "string"]:
            return False
        
        unique_values = current_df[column_name].unique()
        not_na_size = current_df[column_name].notna().size

        # Se os valores únicos são mais de 60% das linhas, não vale a pena
        if (round((unique_values.size / not_na_size) * 100) > 60):
            return False
        
        return True


    @staticmethod
    def to_category(
        current_df: cudf.DataFrame,
        column_name: str
    ) -> None:
        # Coluna sem dados
        if current_df[column_name].notna().sum() == 0:
            return
        
        # A coluna deve ser do tipo str
        if current_df[column_name].dtype not in ["object", "string"]:
            return
        
        unique_values = current_df[column_name].unique()

        unique_values = unique_values.sort_values()
        unique_values = unique_values.reset_index(drop=True)
        
        categorical_dtype = cudf.CategoricalDtype(categories=unique_values, ordered=True)
    
        current_df[column_name] = current_df[column_name].astype(categorical_dtype)
        print_job_create_category_done(column_name)


    @staticmethod
    def is_string(
        current_df: cudf.DataFrame,
        column_name: str
    ) -> None:
        if current_df[column_name].dtype in ["object", "string"]:
            return True
        return False


    @staticmethod
    def to_categories(
        current_df: cudf.DataFrame,
        column_names: list[str]
    ) -> None:
        for column_name in column_names:

            if current_df[column_name].notna().sum() == 0:
                continue
            
            if current_df[column_name].dtype not in ["object", "string"]:
                continue

            unique_values = current_df[column_name].unique()
            not_na_size = current_df[column_name].notna().size

            # Os dados únicos devem ser pelo menos 60% oi mais do total
            if (round((unique_values.size / not_na_size) * 100) > 60):
                continue

            unique_values = unique_values.sort_values()
            unique_values = unique_values.reset_index(drop=True)
            
            categorical_dtype = cudf.CategoricalDtype(categories=unique_values, ordered=True)
        
            current_df[column_name] = current_df[column_name].astype(categorical_dtype)
            print_job_create_category_done(column_name)


    @staticmethod
    def is_probably_text(
        current_df: cudf.DataFrame,
        column_name: str
    ) -> bool:
        """
        Verifica se a coluna mesmo sendo string, tem dados tipicos de outros tipos como
        int, float, bool, datetime, timedelta
        """
        has_a_not_string_data = (
            GPUAnalysisUtils.infer_by_sample(
                series=current_df[column_name],
                regex_patern=MIXED_PATTERN,
                n_parts=100
                )
            )

        # Se contem apenas dados de string, então é uma coluna string real
        return (not has_a_not_string_data)


    @staticmethod
    def normalize_unique_values(
        current_df: cudf.DataFrame,
        column_name: str,
        mapping_dict: dict[str, list[str]],
        null_values: None|list[str] = []
    ) -> None:
        """
        Normaliza e limpa uma coluna em um cuDF DataFrame de forma inplace.

        Args:
            current_df: O DataFrame a ser modificado.
            column_name: O nome da coluna a ser normalizada.
            mapping_dict: Dicionário no formato {'novo_valor': ['antigo1', 'antigo2']}.
            null_values: Lista opcional de valores a serem convertidos para nulo.
        """
        replace_map = {
            old_val: new_val
            for new_val, old_vals_list in mapping_dict.items()
            for old_val in old_vals_list
        }

        # Substitiu os valores para o valor da chave do dict
        current_df[column_name] = current_df[column_name].replace(replace_map)

        # Substitiu os valores nulos
        if null_values:
            current_df[column_name] = current_df[column_name].replace(
                null_values, cudf.NA
            )



import cudf

def print_text_red(text: str) -> str:
    return f"\033[1;31m{text}\033[0m"

def print_text_yellow(text: str) -> str:
    return f"\033[1;33m{text}\033[0m"

def print_text_green(text: str) -> str:
    return f"\033[1;32m{text}\033[0m"


def print_job_create_datetime_column_done(
    column_name: str,
    date_column_name: str,
    time_column_name: str,
) -> None:
    print(
        print_text_green("Done!"),
        "column",
        print_text_yellow(column_name),
        "created from",
        print_text_yellow(date_column_name),
        "and",
        print_text_yellow(time_column_name),
        "type",
        print_text_yellow("datetime64[ns]")
    )


def print_job_create_category_column_done(
    column_name: str,
    column_from: str,
) -> None:
    print(
        print_text_green("Done!"),
        print_text_yellow("category"),
        "column",
        print_text_yellow(column_name),
        "created from",
        print_text_yellow(column_from),
        "type",
        print_text_yellow("category")
    )


class GPUNormalizeDateTimeUtils:
    @staticmethod
    def create_day_week_column(
        current_df: cudf.DataFrame,
        datetime_column_name: str,
        new_day_week_column_name: str = "day_week_i",
        before_column: None|str = None
    ) -> None:
        """
        Cria uma nova coluna datetime somando a coluna date e a coluna time (timedelta).
        """
        if str(current_df[datetime_column_name].dtype) not in ["datetime64[s]", "datetime64[ms]", "datetime64[us]", "datetime64[ns]"]:
            raise TypeError(f"A coluna '{datetime_column_name}' deve estar em datetime.")

        if before_column is not None:
            if before_column not in current_df.columns:
                raise ValueError(f"A coluna '{before_column}' não existe no DataFrame.")
            insert_position: int = current_df.columns.get_loc(before_column)

            current_df.insert(
                loc=insert_position,
                column=new_day_week_column_name,
                value=current_df[datetime_column_name].dt.dayofweek
            )
        else:
            current_df[new_day_week_column_name] = current_df[datetime_column_name].dt.dayofweek


    # @staticmethod
    # def create_day_week_br_category_column(
    #     current_df: cudf.DataFrame,
    #     datetime_column_name: str,
    #     new_day_week_column_name: str = "day_week",
    #     before_column: None | str = None,
    # ) -> None:
    #     """
    #     Cria uma nova coluna com os dias da semana como tipo categórico ordenado.
    #     """
    #     if str(current_df[datetime_column_name].dtype) not in ["datetime64[s]", "datetime64[ms]", "datetime64[us]", "datetime64[ns]"]:
    #         raise TypeError(f"A coluna '{datetime_column_name}' deve estar em datetime.")

    #     # Define a ordem correta dos dias da semana
    #     day_week_categories = [
    #         "segunda",
    #         "terca",
    #         "quarta",
    #         "quinta",
    #         "sexta",
    #         "sabado",
    #         "domingo",
    #     ]
        
    #     # Mapeia os números do dia da semana (0=segunda, 6=domingo) para nomes.
    #     day_week_mapping = {
    #         0: "segunda",
    #         1: "terca",
    #         2: "quarta",
    #         3: "quinta",
    #         4: "sexta",
    #         5: "sabado",
    #         6: "domingo",
    #     }
        
    #     # Obtém a Series com os números do dia da semana e mapeia para os nomes
    #     day_of_week_series = current_df[datetime_column_name].dt.dayofweek.astype("int8")
    #     category_series = day_of_week_series.map(day_week_mapping)
        
    #     # Define o tipo categórico ordenado
    #     ordered_dtype = cudf.CategoricalDtype(categories=day_week_categories, ordered=True)
        
    #     # Converte a série para o novo tipo categórico ordenado
    #     category_series = category_series.astype(ordered_dtype)

    #     if before_column is not None:
    #         if before_column not in current_df.columns:
    #             raise ValueError(f"A coluna '{before_column}' não existe no DataFrame.")
    #         insert_position: int = current_df.columns.get_loc(before_column)
            
    #         current_df.insert(
    #             loc=insert_position,
    #             column=new_day_week_column_name,
    #             value=category_series,
    #         )
    #     else:
    #         current_df[new_day_week_column_name] = category_series
    #     print_job_create_category_column_done(
    #         column_name=new_day_week_column_name,
    #         column_from=datetime_column_name
    #     )


    # @staticmethod
    # def create_month_category_column(
    #     current_df: cudf.DataFrame,
    #     datetime_column_name: str,
    #     new_month_category_column_name: str = "month",
    #     before_column: None | str = None,
    #     case_type: str = "lower"
    # ) -> None:
    #     """
    #     Cria uma nova coluna com os meses do ano como tipo categórico ordenado.
    #     """
    #     if str(current_df[datetime_column_name].dtype) not in ["datetime64[s]", "datetime64[ms]", "datetime64[us]", "datetime64[ns]"]:
    #         raise TypeError(f"A coluna '{datetime_column_name}' deve estar em datetime.")

    #     # Define a ordem correta dos meses do ano
    #     month_categories = [
    #         "janeiro",
    #         "fevereiro",
    #         "marco",
    #         "abril",
    #         "maio",
    #         "junho",
    #         "julho",
    #         "agosto",
    #         "setembro",
    #         "outubro",
    #         "novembro",
    #         "dezembro",
    #     ]
        
    #     # Mapeia os números do mês (1=janeiro, 12=dezembro) para nomes.
    #     month_mapping = {
    #         1: "janeiro",
    #         2: "fevereiro",
    #         3: "marco",
    #         4: "abril",
    #         5: "maio",
    #         6: "junho",
    #         7: "julho",
    #         8: "agosto",
    #         9: "setembro",
    #         10: "outubro",
    #         11: "novembro",
    #         12: "dezembro",
    #     }
        
    #     # Obtém a Series com os números dos meses e mapeia para os nomes
    #     month_series = current_df[datetime_column_name].dt.month
    #     category_series = month_series.map(month_mapping)
        
    #     # Define o tipo categórico ordenado
    #     ordered_dtype = cudf.CategoricalDtype(categories=month_categories, ordered=True)
        
    #     # Converte a série para o novo tipo categórico ordenado
    #     category_series = category_series.astype(ordered_dtype)

    #     if before_column is not None:
    #         if before_column not in current_df.columns:
    #             raise ValueError(f"A coluna '{before_column}' não existe no DataFrame.")
    #         insert_position: int = current_df.columns.get_loc(before_column)
            
    #         current_df.insert(
    #             loc=insert_position,
    #             column=new_month_category_column_name,
    #             value=category_series,
    #         )
    #     else:
    #         current_df[new_month_category_column_name] = category_series

    #     print_job_create_category_column_done(
    #         column_name=new_month_category_column_name,
    #         column_from=datetime_column_name
    #     )

    @staticmethod
    def create_day_week_br_category_column(
        current_df: cudf.DataFrame,
        datetime_column_name: str,
        new_day_week_column_name: str = "day_week",
        before_column: None | str = None,
        case_type: str = "upper",
    ) -> None:
        """
        Cria uma nova coluna com os dias da semana como tipo categórico ordenado,
        com a opção de formatar o case dos nomes dos dias da semana.
        """
        if str(current_df[datetime_column_name].dtype) not in ["datetime64[s]", "datetime64[ms]", "datetime64[us]", "datetime64[ns]"]:
            raise TypeError(f"A coluna '{datetime_column_name}' deve estar em datetime.")

        # Lista de dias da semana em minúsculas como base
        day_week_categories_base = [
            "SEGUNDA", "TERCA", "QUARTA", "QUINTA", "SEXTA", "SABADO", "DOMINGO",
        ]
        
        # Mapeia os números do dia da semana (0=segunda, 6=domingo) para nomes.
        day_week_mapping = {i: day for i, day in enumerate(day_week_categories_base)}
        
        # Obtém a série com os números do dia da semana e mapeia para os nomes
        day_of_week_series = current_df[datetime_column_name].dt.dayofweek
        category_series = day_of_week_series.map(day_week_mapping)
        
        # Aplica a formatação do case
        if case_type == "lower":
            day_week_categories = [d.lower() for d in day_week_categories_base]
            category_series = category_series.str.lower()

        elif case_type == "capitalize":
            day_week_categories = [d.capitalize() for d in day_week_categories_base]
            category_series = category_series.str.capitalize()

        elif case_type == "upper":
            day_week_categories = day_week_categories_base
            # Não precisa de conversão, pois a base já está em minúsculas
        else:
            raise ValueError("O parâmetro 'case_type' deve ser 'upper', 'lower' ou 'capitalize'.")
        
        # Define o tipo categórico ordenado
        ordered_dtype = cudf.CategoricalDtype(categories=day_week_categories, ordered=True)
        
        # Converte a série para o novo tipo categórico ordenado
        category_series = category_series.astype(ordered_dtype)

        if before_column is not None:
            if before_column not in current_df.columns:
                raise ValueError(f"A coluna '{before_column}' não existe no DataFrame.")
            insert_position: int = current_df.columns.get_loc(before_column)
            
            current_df.insert(
                loc=insert_position,
                column=new_day_week_column_name,
                value=category_series,
            )
        else:
            current_df[new_day_week_column_name] = category_series


    @staticmethod
    def create_month_br_category_column(
        current_df: cudf.DataFrame,
        datetime_column_name: str,
        new_month_category_column_name: str = "month",
        before_column: None | str = None,
        case_type: str = "upper",
    ) -> None:
        """
        Cria uma nova coluna com os meses do ano como tipo categórico ordenado,
        com a opção de formatar o case dos nomes dos meses.
        """
        if str(current_df[datetime_column_name].dtype) not in ["datetime64[s]", "datetime64[ms]", "datetime64[us]", "datetime64[ns]"]:
            raise TypeError(f"A coluna '{datetime_column_name}' deve estar em datetime.")

        month_categories_base = [
            "JANEIRO", "FEVEREIRO", "MARCO", "ABRIL", "MAIO", "JUNHO",
            "JULHO", "AGOSTO", "SETEMBRO", "OUTUBRO", "NOVEMBRO", "DEZEMBRO",
        ]

        month_mapping = {i + 1: month for i, month in enumerate(month_categories_base)}

        if case_type == "lower":
            month_categories = [m.lower() for m in month_categories_base]
            category_series = current_df[datetime_column_name].dt.month.map(month_mapping).str.lower()

        elif case_type == "capitalize":
            month_categories = [m.capitalize() for m in month_categories_base]
            category_series = current_df[datetime_column_name].dt.month.map(month_mapping).str.capitalize()

        elif case_type == "upper":
            month_categories = month_categories_base
            category_series = current_df[datetime_column_name].dt.month.map(month_mapping)

        else:
            raise ValueError("O parâmetro 'case_type' deve ser 'upper', 'lower' ou 'capitalize'.")

        # Define o tipo categórico ordenado com os nomes formatados
        ordered_dtype = cudf.CategoricalDtype(categories=month_categories, ordered=True)
        
        # Converte a série para o novo tipo categórico ordenado
        category_series = category_series.astype(ordered_dtype)

        # Lógica de inserção da coluna
        if before_column is not None:
            if before_column not in current_df.columns:
                raise ValueError(f"A coluna '{before_column}' não existe no DataFrame.")
            insert_position: int = current_df.columns.get_loc(before_column)
            current_df.insert(loc=insert_position, column=new_month_category_column_name, value=category_series)
        else:
            current_df[new_month_category_column_name] = category_series

        print_job_create_category_column_done(
            column_name=new_month_category_column_name,
            column_from=datetime_column_name
        )



    @staticmethod
    def create_datetime_column_from_date_and_time(
        current_df: cudf.DataFrame,
        date_column_name: str,
        time_column_name: str,
        new_datetime_column_name: str = "datetime",
        before_column: None|str = None
    ) -> None:
        """
        Cria uma nova coluna datetime somando a coluna date e a coluna time (timedelta).
        """
        if str(current_df[date_column_name].dtype) not in ["datetime64[s]", "datetime64[ms]", "datetime64[us]", "datetime64[ns]"]:
            raise TypeError(f"A coluna '{date_column_name}' deve estar em datetime.")

        if str(current_df[time_column_name].dtype) not in ["timedelta64[s]", "timedelta[ms]", "timedelta64[ms]", "timedelta[us]",  "timedelta64[us]", "timedelta[ns]", "timedelta64[ns]"]:
            raise TypeError(f"A coluna '{time_column_name}' deve estar em timedelta.")
        
        new_column_values = current_df[date_column_name] + current_df[time_column_name]

        if before_column is not None:
            if before_column not in current_df.columns:
                raise ValueError(f"A coluna '{before_column}' não existe no DataFrame.")
            insert_position: int = current_df.columns.get_loc(before_column)

            current_df.insert(
                loc=insert_position,
                column=new_datetime_column_name,
                value=new_column_values
            )
        else:
            current_df[new_datetime_column_name] = new_column_values

        print_job_create_datetime_column_done(
            column_name=new_datetime_column_name,
            date_column_name=date_column_name,
            time_column_name=time_column_name,
        )


    @staticmethod
    def create_hour_category_column(
        current_df: cudf.DataFrame,
        datetime_column_name: str,
        new_hour_category_column_name: str = "hour",
        before_column: None | str = None,
    ) -> None:
        """
        Cria uma nova coluna com a hora do dia (00 a 23) como tipo categórico ordenado.
        """
        if str(current_df[datetime_column_name].dtype) not in ["datetime64[s]", "datetime64[ms]", "datetime64[us]", "datetime64[ns]"]:
            raise TypeError(f"A coluna '{datetime_column_name}' deve estar em datetime.")

        # Define a ordem correta das horas do dia (00 a 23)
        # A lista de categorias garante a ordem numérica correta.
        hour_categories = [f"{i:02d}" for i in range(24)]
        
        # Obtém a série com os números das horas (0 a 23)
        hour_series = current_df[datetime_column_name].dt.hour
        
        # Converte os números das horas para strings com zero à esquerda (ex: 5 -> "05")
        # Isso é necessário para o formato "HH" e para o tipo categórico.
        hour_string_series = hour_series.astype(str).str.zfill(2)
        
        # Define o tipo categórico ordenado
        ordered_dtype = cudf.CategoricalDtype(categories=hour_categories, ordered=True)
        
        # Converte a série de strings para o novo tipo categórico ordenado
        category_series = hour_string_series.astype(ordered_dtype)

        if before_column is not None:
            if before_column not in current_df.columns:
                raise ValueError(f"A coluna '{before_column}' não existe no DataFrame.")
            insert_position: int = current_df.columns.get_loc(before_column)
            
            current_df.insert(
                loc=insert_position,
                column=new_hour_category_column_name,
                value=category_series,
            )
        else:
            current_df[new_hour_category_column_name] = category_series

        print_job_create_category_column_done(
            column_name=new_hour_category_column_name,
            column_from=datetime_column_name
        )
